import os
import re
import subprocess


def available_cpu_count():
    """ Number of available virtual or physical CPUs on this system, i.e.
    user/real as output by time(1) when called with an optimally scaling
    userspace-only program

    Inspired by: https://stackoverflow.com/a/1006301
    """

    # cpuset may restrict the number of *available* processors
    try:
        with open('/proc/self/status') as fh:
            m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$', fh.read())
        if m:
            res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
            if res > 0:
                return res
    except IOError:
        pass

    # Python 2.6+
    try:
        import multiprocessing
        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    # https://github.com/giampaolo/psutil
    try:
        import psutil
        return psutil.cpu_count()  # psutil.NUM_CPUS on old versions
    except (ImportError, AttributeError):
        pass

    # POSIX
    try:
        res = int(os.sysconf('SC_NPROCESSORS_ONLN'))

        if res > 0:
            return res
    except (AttributeError, ValueError):
        pass

    # Windows
    try:
        res = int(os.environ['NUMBER_OF_PROCESSORS'])

        if res > 0:
            return res
    except (KeyError, ValueError):
        pass

    # jython
    try:
        from java.lang import Runtime
        runtime = Runtime.getRuntime()
        res = runtime.availableProcessors()
        if res > 0:
            return res
    except ImportError:
        pass

    # BSD
    try:
        sysctl = subprocess.Popen(['sysctl', '-n', 'hw.ncpu'], stdout=subprocess.PIPE)
        sc_stdout = sysctl.communicate()[0]
        res = int(sc_stdout)

        if res > 0:
            return res
    except (OSError, ValueError):
        pass

    # Linux
    try:
        with open('/proc/cpuinfo') as fh:
            res = fh.read().count('processor\t:')

        if res > 0:
            return res
    except IOError:
        pass

    # Solaris
    try:
        pseudo_devices = os.listdir('/devices/pseudo/')
        res = 0
        for pd in pseudo_devices:
            if re.match(r'^cpuid@[0-9]+$', pd):
                res += 1

        if res > 0:
            return res
    except OSError:
        pass

    # Other UNIXes (heuristic)
    try:
        try:
            with open('/var/run/dmesg.boot') as fh:
                dmesg = fh.read()
        except IOError:
            dmesg_process = subprocess.Popen(['dmesg'], stdout=subprocess.PIPE)
            dmesg = dmesg_process.communicate()[0]

        res = 0
        while '\ncpu' + str(res) + ':' in dmesg:
            res += 1

        if res > 0:
            return res
    except OSError:
        pass

    raise Exception('Can not determine the number of CPU(s) on this system')



import json
import pathlib
import random
import shutil
import uuid
import warnings
from io import BytesIO, StringIO
from itertools import groupby
from typing import Optional, Dict

import cv2
import imagesize
import numpy as np
import requests
from PIL import Image
from rays_pycocotools import coco
from rays_pycocotools import mask as pycoco_mask
from tqdm import tqdm

import asyncio
from timeit import default_timer
from concurrent.futures import ThreadPoolExecutor

from raytils.datasets.coco_fields import CocoCategories, CocoInfo, CocoLicence, CocoDetectionAnnotation, CocoModes, \
    CocoList, CocoImage, combine_coco_objects
from raytils.perception import get_image_size


class Annotation:
    def __init__(self, class_name, bbox=None, mask=None, classifications=None, kwargs=None):
        if classifications is None:
            classifications = []
        if kwargs is None:
            kwargs = {}
        if bbox is None and mask is None:
            raise ValueError("No annotations provided either bbox or mask must be valid")
        if not isinstance(classifications, list):
            raise TypeError("Classifications parameter must be a list!")
        if not isinstance(class_name, str):
            raise TypeError("Class name parameter must be a string!")
        self.class_name = class_name
        self.classifications = classifications
        self.kwargs = kwargs
        self._bbox = bbox
        self._mask = mask
        self.__bbox_area = 0
        self.__mask_area = 0
        _, _, self._bbox, self._mask = self.valid()  # Force numpy arrays or large memory hogs to lowest representation


    def valid(self):
        bbox, poly = self.bbox, self.poly
        valid_bbox, valid_poly = sum([x == 0 for x in bbox]) != len(bbox), poly and bool(poly[0])
        return valid_bbox, valid_poly, bbox, poly

    @property
    def area(self):
        _, _ = self.poly, self.bbox  # Getters may set area
        if self.__mask_area:
            return self.__mask_area
        if self.__bbox_area:
            return self.__bbox_area
        return 0

    @property
    def poly(self):
        # TODO: Find general purpose way of doing this (mask->segm, RLE->segm, RLEencoded->segm)
        # TODO: Currently only supports RLEencoded->segm
        if self._mask is None:  # Return None if no poly
            return self._mask

        if isinstance(self._mask, list) and any(isinstance(el, list) for el in self._mask):
            return self._mask  # Must already be a list of poly lists
        elif isinstance(self._mask, (np.ndarray, dict)):  # If mask or RLE then get contours
            if isinstance(self._mask, dict):  # RLE
                if isinstance(self._mask['counts'], (bytes, list)):
                    # Decode RLE to mask
                    if isinstance(self._mask['counts'], list):
                        rle = coco.maskUtils.frPyObjects(self._mask, self._mask["size"][0], self._mask["size"][1])
                    else:
                        rle = self._mask
                    mask = coco.maskUtils.decode(rle)

                    if self._bbox is None:
                        self._bbox = list(coco.maskUtils.toBbox(rle))
                else:
                    raise ValueError("Mask is not a valid RLE dict")
            else:  # Must already be mask?
                mask = self._mask

            # Decode the binary mask
            contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
            segmentation = [[]]
            if contours:
                segmentation = [c.flatten().tolist() for c in contours if c.size > 4]

            if self._bbox is None:
                xfl = [i for j in segmentation for idx, i in enumerate(j) if idx % 2 == 0]
                yfl = [i for j in segmentation for idx, i in enumerate(j) if idx % 2 != 0]
                x, y = min(xfl), min(yfl)
                w, h = max(xfl) - x, max(yfl) - y
                self._bbox = [x, y, w, h]

            self.__mask_area = np.count_nonzero(mask)
            self._mask = segmentation  # Change mask to poly to save memory
            return self._mask
        elif isinstance(self._mask, list) and not isinstance(self._mask[0], list) and len(self._mask) % 2 == 0:
            self._mask = [self._mask]  # Mask must already be a poly
            return self._mask

        raise ValueError("Invalid value for mask ({})".format(type(self._mask)))

    @property
    def bbox(self):
        if self._bbox:
            self.__bbox_area = self._bbox[-2] * self._bbox[-1]
            return self._bbox

        _ = self.poly  # Poly may set the bounding box

        if self._bbox is None:
            raise ValueError("Cannot automatically infer bounding box")

        return self._bbox


class COCOCreator:
    """ Class for creating annotation files in COCO format through a simple interface

    Attributes:
        mode: Coco annotation file type e.g "detection" or "keypoints"
        data_splits: Dictionary describing the export format (useful for splitting into training/testing splits)
        random_splits: Shuffle the imported data between data_splits if true
    """

    # TODO: Finish implementation
    #  - Add support for multiple different file imports (csv, json, images etc)
    #  - Add support for different modes (Keypoint, Panoptic, Instance Segm etc)
    def __init__(self, mode: str, data_splits: Optional[Dict[str, float]] = None, random_splits=True):
        if not CocoModes.is_valid(mode):
            raise ValueError(f"Mode '{mode}' isn't a valid COCO format\n\tMust be one of: {CocoModes.modes}")
        self.__mode = mode
        if self.__mode != CocoModes.DETECTION:
            raise NotImplementedError(f"Export of {self.__mode} coco files is not currently implemented")
        if data_splits is None:
            data_splits = {"coco": 1.0}
        if 0 > sum(data_splits.values()) > 1.0:
            raise ValueError(f"Cannot split dataset into '{data_splits}' data splits are not between 0 and 1!")
        self.data_splits = data_splits
        self.random_splits = random_splits
        self.__image_path_ann_lut = {}
        self.__extra_fields_lut = {}
        self.classes = CocoCategories()
        self.info = CocoInfo(contributor="Raymond Kirk", url="https://github.com/RaymondKirk")
        self.license = CocoLicence(licence_id=0, name="Contact Author", url="https://github.com/RaymondKirk")
        self.licenses = CocoList([self.license])

    def map_classes(self, map_dict):
        """Replace classes in keys of map_dict with value"""
        for key, annotations in self.__image_path_ann_lut.items():
            for ann in annotations:
                if ann.class_name in map_dict:
                    ann.class_name = map_dict[ann.class_name]
        for old, new in map_dict.items():
            if old in self.classes:
                self.classes.remove(old)
            else:
                warnings.warn(f"Class '{old}' does not exist. Adding new class '{new}' anyway.")
            self.classes.add(new)

    def remove_classes(self, classes):
        """Replace classes in keys of map_dict with value"""
        for key in self.__image_path_ann_lut.keys():
            self.__image_path_ann_lut[key] = [x for x in self.__image_path_ann_lut[key] if x.class_name not in classes]

        for old in classes:
            self.classes.remove(old)

    def add_classes(self, classes):
        if not isinstance(classes, list):
            classes = [classes]
        for cls in classes:
            self.classes.add(cls)

    def add_detection_annotation(self, image_path, class_name, bbox=None, mask=None, classifications=None, kwargs=None):
        if self.__mode != CocoModes.DETECTION:
            raise ValueError(f"Cannot call add_detection_annotation when creating a {self.__mode} coco file")
        image_path = str(image_path)  # Ensure it's a string not Path object
        if image_path not in self.__image_path_ann_lut:
            self.__image_path_ann_lut[image_path] = []
        annotation = Annotation(class_name, bbox, mask, classifications, kwargs)
        self.__image_path_ann_lut[image_path].append(annotation)
        self.classes.add(class_name)

    def add_detection_image(self, image_path):
        if self.__mode != CocoModes.DETECTION:
            raise ValueError(f"Cannot call add_detection_annotation when creating a {self.__mode} coco file")
        image_path = str(image_path)  # Ensure it's a string not Path object
        if image_path not in self.__image_path_ann_lut:
            self.__image_path_ann_lut[image_path] = []

    def add_images_with_masks(self, image_list, mask_list, classes_list):
        """Adds detections from an image and corresponding binary mask"""
        for image_path, mask_path, mask_class in zip(image_list, mask_list, classes_list):
            image_path, mask_path = pathlib.Path(image_path), pathlib.Path(mask_path)
            assert image_path.name == mask_path.name
            mask = cv2.imread(str(mask_path), cv2.IMREAD_GRAYSCALE)
            img = cv2.imread(str(image_path), cv2.IMREAD_COLOR)
            if mask.shape[:2] != img.shape[:2]:
                mask = cv2.resize(mask, (img.shape[:2])[::-1])
            self.add_detection_annotation(image_path, mask_class, None, mask)

    def add_images_with_greyscale_mask(self, image_list, mask_list, classes_list):
        """Adds detections from an image and corresponding greyscale mask (multiple annotations to an image)"""
        for image_path, mask_path, mask_class in tqdm(total=zip(image_list, mask_list, classes_list),
                                                      postfix="Adding images from segmentation masks"):
            image_path, mask_path = pathlib.Path(image_path), pathlib.Path(mask_path)
            assert image_path.name == mask_path.name
            masks = cv2.imread(str(mask_path), cv2.IMREAD_GRAYSCALE)
            img = cv2.imread(str(image_path), cv2.IMREAD_COLOR)
            if masks.shape[:2] != img.shape[:2]:
                masks = cv2.resize(masks, (img.shape[:2])[::-1])
            unique_values = np.unique(masks)
            unique_values = unique_values[unique_values != 0].tolist()
            for mask_idx in unique_values:
                mask = np.zeros_like(masks)
                mask[np.where(masks == mask_idx)] = 255.0
                self.add_detection_annotation(image_path, mask_class, None, mask)

    def add_coco_json(self, json_path, image_root=None):
        """Adds detections from a coco file"""
        if isinstance(json_path, str):
            json_path = pathlib.Path(json_path)
        if image_root is None:
            image_root = json_path.parent
        with json_path.open("rb") as fh:
            coco_data = json.load(fh)

        image_id_to_filename = {c["id"]: c["file_name"] for c in coco_data["images"]}
        category_id_to_class_name = {c["id"]: c["name"] for c in coco_data["categories"]}
        for annotation in coco_data["annotations"]:
            bbox = annotation.get("bbox")
            segm = annotation.get("segmentation")
            class_name = category_id_to_class_name[annotation["category_id"]]
            file_name = image_root / image_id_to_filename[annotation["image_id"]]
            self.add_detection_annotation(file_name, class_name, bbox, segm)

    def add_darwin_v7(self, json_path):
        """Adds detections from a V7 darwin json file"""
        if isinstance(json_path, str):
            json_path = pathlib.Path(json_path)
        with json_path.open("rb") as fh:
            darwin = json.load(fh)

        file_id_to_name = {str(idx): d for idx, d in enumerate(darwin["image"]["frame_urls"])}
        for path in file_id_to_name.values():  # add all images in-case of no annotations for image
            self.add_detection_image(path)

        annotations = {}  # frame_id: [annotations]
        for ann in darwin["annotations"]:
            class_name = ann["name"]
            interpolated = ann["interpolated"]
            # validate all instance_ids are consistent
            ids = set([ii["instance_id"]["value"] for ii in ann["frames"].values() if "instance_id" in ii])
            if not ids:
                continue
            if len(ids) > 1:
                raise ValueError("Instance IDS not unique")
            instance_id = ids.pop()
            for frame_id, frame_annotation in ann["frames"].items():
                bbox = [frame_annotation["bounding_box"].get(a) for a in ["x", "y", "w", "h"]]
                if frame_id not in annotations:
                    annotations[frame_id] = []
                self.add_detection_annotation(image_path=file_id_to_name[frame_id], bbox=bbox, class_name=class_name,
                                              kwargs=dict(instance_id=instance_id, interpolated=interpolated,))

    def add_label_box_export_async(self, json_path, num_workers=64):
        """Converts a LabelBox JSON (https://app.labelbox.com/) instance segmentation file to COCO"""
        async def get_label_box_async(json_file_path):
            def fetch(requests_session, item):
                rgb_file = item["Labeled Data"]
                objects = item["Label"]["objects"]
                max_attempts = 20
                status = "Done"

                for o in objects:
                    mask_url = o["instanceURI"]
                    attempts = 0
                    while attempts <= max_attempts:
                        r = requests_session.get(mask_url)
                        if r.status_code != 200:
                            print(
                                f"\nError code {r.status_code}={r.reason} for {mask_url} (attempt={attempts}/{max_attempts})")
                            if attempts == max_attempts:
                                print(f"Skipping {rgb_file}, too many failed attempts")
                                status = "Error"
                                break
                            attempts += 1
                            continue
                        mask = np.asfortranarray(Image.open(BytesIO(r.content)), dtype=np.uint8)[..., 0]
                        mask[mask > 0] = 1.0
                        rle = pycoco_mask.encode(mask)
                        classifications = [a["value"] for i in o.get("classifications", []) for a in i["answers"]]
                        self.add_detection_annotation(rgb_file, o["value"], mask=rle, classifications=classifications)
                        break

                elapsed = default_timer() - async_loop_start_time
                time_completed_at = "{:5.2f}s".format(elapsed)
                progress_bar.set_postfix_str("{} with {} at {}".format(status, rgb_file, time_completed_at))
                progress_bar.update(1)

            if isinstance(json_file_path, str):
                json_file_path = pathlib.Path(json_file_path).resolve()

            with json_file_path.open("r") as fh:
                json_data = json.load(fh)

            # Initialize the event loop
            loop = asyncio.get_event_loop()
            with ThreadPoolExecutor(max_workers=num_workers) as executor:
                with requests.Session() as session:
                    progress_bar = tqdm(total=len(json_data))
                    async_loop_start_time = default_timer()

                    tasks = [
                        loop.run_in_executor(
                            executor,
                            fetch,
                            *(session, item)
                        )
                        for item in json_data
                    ]

                    # Initializes the tasks to run and awaits their results
                    for response in await asyncio.gather(*tasks):
                        pass

                    progress_bar.close()

        loop = asyncio.get_event_loop()
        future = asyncio.ensure_future(get_label_box_async(json_path))
        loop.run_until_complete(future)

    def add_label_me_folder(self, folder_path):
        """Converts a LabelMe (https://github.com/wkentaro/labelme) instance segmentation folder to COCO"""
        if isinstance(folder_path, str):
            folder_path = pathlib.Path(folder_path).resolve()

        json_files = list(folder_path.glob("*.json"))

        if not len(json_files):
            raise IOError("No '.json' files in the label me folder '{}'!".format(folder_path))

        for file in tqdm(total=json_files, postfix=f"Detections from LabelMe Folder {folder_path}"):
            with file.open("rb") as fh:
                label_data = json.load(fh)
            image_path = (folder_path / label_data["imagePath"]).resolve()
            if not image_path.is_file():
                raise IOError("File '{}' does not exists!".format(image_path))

            for shape in label_data['shapes']:
                if "points" not in shape:
                    raise NotImplementedError("Only LabelMe files labelled as polygons are supported")
                points_list = [item for xy in shape['points'] for item in xy]
                x1, x2 = min([xy[0] for xy in shape['points']]), max([xy[0] for xy in shape['points']])
                y1, y2 = min([xy[1] for xy in shape['points']]), max([xy[1] for xy in shape['points']])
                w, h = x2 - x1, y2 - y1
                label_no_digits = ''.join(i for i in shape["label"] if not i.isdigit())
                # TODO: Calculate mask area instead of bbox
                self.add_detection_annotation(image_path, label_no_digits, bbox=[x1, y1, w, h], mask=points_list)

    def add_extra_info(self, image_path, **kwargs):
        """Adds extra info to a CocoImage field. Useful for things such as depth.
            If adding extra file paths you must also copy the files before or after exporting.
        """
        if image_path not in self.__image_path_ann_lut:
            raise ValueError(f"Annotations for '{image_path}' have not been added yet. Please add them first.")

        if image_path not in self.__extra_fields_lut:
            self.__extra_fields_lut[image_path] = {}
        for field_name, field_info in kwargs.items():
            self.__extra_fields_lut[image_path][field_name] = field_info

    def export(self, dataset_root: pathlib.Path, *args, **kwargs):
        if self.__mode == CocoModes.DETECTION:
            return self.export_detection(dataset_root, *args, **kwargs)
        raise NotImplementedError(f"Mode  '{self.__mode}' is not currently supported.")

    def export_detection(self, output_dir: pathlib.Path, symlink_to_image: bool = False, exists_ok: bool = False):
        if isinstance(output_dir, str):
            output_dir = pathlib.Path(output_dir)
        output_dir.mkdir(exist_ok=True, parents=True)
        class_name_to_idx = self.classes.classes_to_idx

        image_id = 0

        image_path_lut = self.__image_path_ann_lut
        packaged_image_path_lut = {}
        progress_bar_post_fix = ("Creating symlink to" if symlink_to_image else "Copying") + " image data '{}'"
        progress_bar = tqdm(total=len(image_path_lut), postfix=progress_bar_post_fix.format(""))

        images_not_local = False
        tmp_dir = f"/tmp/coco_creator/{uuid.uuid4()}"

        # Write image data to new packaged dataset
        for image_file, image_annotation in image_path_lut.items():
            original_file = image_file
            # Download the file if it doesn't exist locally
            if "http://" in image_file or "https://" in image_file:
                images_not_local = True
                tmp_file = pathlib.Path(f"{tmp_dir}/{pathlib.Path(image_file).name}")
                tmp_file.parent.mkdir(exist_ok=True)
                with tmp_file.open("wb") as handle:
                    response = requests.get(image_file, stream=True)
                    if not response.ok:
                        raise IOError("Could not download image_file")
                    progress_bar.set_postfix_str(f"Downloading {image_file}")
                    for block in response.iter_content(1024):
                        if not block:
                            break
                        handle.write(block)
                image_file = type(original_file)(tmp_file)

            if not pathlib.Path(image_file).exists():
                print(f"Skipping '{image_file}' as it does not exist on the filesystem")
                continue
            progress_bar.set_postfix_str(progress_bar_post_fix.format(image_file))

            width, height = get_image_size(image_file)

            # Copy the files to new COCO dataset
            new_image_file = f"data/{image_id}_{pathlib.Path(image_file).name}"
            new_image_file_location = output_dir / new_image_file
            new_image_file_location.parent.mkdir(exist_ok=True, parents=True)
            if symlink_to_image:  # Symlink if parameter is set else copy the file
                if images_not_local:
                    raise IOError("Cannot symlink to images if they're not stored locally")
                try:
                    pathlib.Path(new_image_file_location).symlink_to(pathlib.Path(image_file),
                                                                     target_is_directory=False)
                except FileExistsError:
                    if not exists_ok:  # sometimes if combining datasets the file will exist in multiple coco files
                        raise
            else:
                shutil.copy(str(image_file), str(new_image_file_location))

            packaged_image_path_lut[new_image_file] = (CocoImage(
                image_id=image_id,
                width=width,
                height=height,
                file_name=new_image_file,
                image_license=self.license.id,
                flickr_url="",
                coco_url=original_file,
                date_captured="",
                **self.__extra_fields_lut.get(image_file, {})
            ), image_path_lut[original_file])

            image_id += 1
            progress_bar.update()
        progress_bar.close()
        del image_path_lut

        # Split files randomly into n-sized data split sections
        data_split_keys = list(self.data_splits.keys())
        data_split_values = [self.data_splits[k] for k in data_split_keys]
        split_sizes = [int(round(len(packaged_image_path_lut) * v)) for v in data_split_values]
        idx_splits = [0] + [sum(split_sizes[:l]) for l in range(1, len(split_sizes) + 1)]
        file_path_keys = list(packaged_image_path_lut.keys())
        if self.random_splits:
            random.shuffle(file_path_keys)
        # Always export an _combined.json file unless this is what the user is exporting/only one file exported
        all_data_splits_file = '_'.join(self.data_splits.keys()) + '_combined'
        if len(self.data_splits.keys()) > 1 and all_data_splits_file not in self.data_splits:
            split_keys = {all_data_splits_file: file_path_keys}
        else:
            split_keys = dict()
        split_keys.update({data_split_keys[i]: file_path_keys[idx_splits[i]:idx_splits[i + 1]]
                           for i in range(0, len(idx_splits) - 1)})

        # Write COCO image (training/testing/all etc splits)
        for split_name, file_paths in split_keys.items():
            annotation_id = 0
            annotations = CocoList()
            export_file_name = output_dir / "{}.json".format(split_name)
            export_file_name.parent.mkdir(parents=True, exist_ok=True)

            progress_bar_post_fix = f" Exporting {export_file_name}: " + "'{}'"
            progress_bar = tqdm(total=len(packaged_image_path_lut), postfix=progress_bar_post_fix.format(""))
            coco_images = CocoList()
            split_packaged_lut = {k: packaged_image_path_lut[k] for k in file_paths}

            for image_file, (coco_image, image_annotation) in split_packaged_lut.items():
                coco_images.append(coco_image)
                for idx, ann in enumerate(image_annotation):
                    progress_bar.set_postfix_str(
                        progress_bar_post_fix.format(f"{idx + 1}/{len(image_annotation)} {image_file}")
                    )
                    valid_bbox, valid_poly, bbox, poly = ann.valid()
                    if not valid_bbox:  # mini
                        continue
                    annotations.append(CocoDetectionAnnotation(
                        annotation_id=annotation_id,
                        image_id=coco_image.id,
                        category_id=class_name_to_idx[ann.class_name],
                        segmentation=poly,
                        area=ann.area,
                        bbox=bbox,
                        iscrowd=0,
                        classifications=ann.classifications,
                        **ann.kwargs
                    ))
                    annotation_id += 1
                progress_bar.update()
            progress_bar.close()
            coco_file = combine_coco_objects([self.info, coco_images, annotations, self.classes, self.licenses])

            with export_file_name.open('w') as fh:
                json.dump(coco_file, fh, indent=4, sort_keys=True)



from typing import Union

import numpy as np
import scipy.spatial

__all__ = ["scale_polygon", "get_points_perimeter", "get_polygon_perimeter", "get_polygon_area", "get_point_along_line"]

from scipy.spatial import distance


def scale_polygon(polygon: np.array, shrink_around: np.array, scale: float, unique: bool = False,
        preserve_type: bool = False) -> np.array:
    """ Takes a polygon and a point and returns a shrunken/grown polygon

    Args:
        polygon: (n, 2) array of xy locations for a closed polygon
        shrink_around: np.array (N, 2) or (1, 2) set of X, Y points to shrink around
        scale: Shrink amount

    Returns:
        Polygon scaled by the percentage scale (between 0 and 1)
    """
    if scale == 1.0:
        return polygon
    if scale == 0.0:
        scale = np.finfo(float).eps
    scaled_polygon = (polygon + ((shrink_around - polygon) * scale))
    if preserve_type:
        scaled_polygon = scaled_polygon.astype(polygon.dtype)
    if unique:
        unique_values, unique_indexes = np.unique(scaled_polygon, axis=0, return_index=True)
        scaled_polygon = scaled_polygon[unique_indexes]
    return scaled_polygon


def get_points_perimeter(x: Union[np.array, list], y: Union[np.array, list]) -> float:
    """ Returns the polygon perimeter for 2D set of (x, y) points

    Args:
        x: (n) Numpy array of x locations for a closed polygon
        y: (n) Numpy array of y locations for a closed polygon

    Returns:
        Perimeter of the closed polygon defined by x, y
    """
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


def get_polygon_perimeter(polygon: np.array) -> float:
    """ Returns the polygon area (perimeter) for 2D set of (x, y) points

    Args:
        polygon: (n, 2) array of xy locations for a closed polygon

    Returns:
        Perimeter of the closed polygon
    """
    return get_points_perimeter(polygon[:, 0], polygon[:, 1])


# unit normal vector of plane defined by points a, b, and c
def unit_normal(a, b, c):
    x = np.linalg.det([[1, a[1], a[2]],
                       [1, b[1], b[2]],
                       [1, c[1], c[2]]])
    y = np.linalg.det([[a[0], 1, a[2]],
                       [b[0], 1, b[2]],
                       [c[0], 1, c[2]]])
    z = np.linalg.det([[a[0], a[1], 1],
                       [b[0], b[1], 1],
                       [c[0], c[1], 1]])
    magnitude = (x ** 2 + y ** 2 + z ** 2) ** .5
    return (x / magnitude, y / magnitude, z / magnitude)


# area of polygon poly
def poly_area(poly):
    if len(poly) < 3:  # not a plane - no area
        return 0
    total = [0, 0, 0]
    N = len(poly)
    for i in range(N):
        vi1 = poly[i]
        vi2 = poly[(i + 1) % N]
        prod = np.cross(vi1, vi2)
        total[0] += prod[0]
        total[1] += prod[1]
        total[2] += prod[2]
    result = np.dot(total, unit_normal(poly[0], poly[1], poly[2]))
    return abs(result / 2)


def sortpts_clockwise(A):
    # Sort A based on Y(col-2) coordinates
    sortedAc2 = A[np.argsort(A[:, 1]), :]

    # Get top two and bottom two points
    top2 = sortedAc2[0:2, :]
    bottom2 = sortedAc2[2:, :]

    # Sort top2 points to have the first row as the top-left one
    sortedtop2c1 = top2[np.argsort(top2[:, 0]), :]
    top_left = sortedtop2c1[0, :]

    # Use top left point as pivot & calculate sq-euclidean dist against
    # bottom2 points & thus get bottom-right, bottom-left sequentially
    sqdists = distance.cdist(top_left[None], bottom2, 'sqeuclidean')
    rest2 = bottom2[np.argsort(np.max(sqdists, 0))[::-1], :]

    # Concatenate all these points for the final output
    return np.concatenate((sortedtop2c1, rest2), axis=0)


def get_polygon_area_experimental(polygon: np.array) -> float:
    """ Returns the polygon area (perimeter) for 2D set of (x, y) points

    Args:
        polygon: (n, 2) array of xy locations for a closed polygon

    Returns:
        Area of the closed polygon
    """
    import warnings
    warnings.warn("WARNING: TESTS FOR GET_POLYGON_AREA NOT CURRENTLY PASSING USE WITH CAUTION")

    if polygon.shape[0] < 3:
        return 0
    # Due to being in 2D area refers to perimeter and volume to area
    polygon = sortpts_clockwise(polygon)
    polygon = np.c_[polygon, np.ones(polygon.shape[0])]
    return poly_area(polygon.tolist())


def get_polygon_area(polygon: np.array) -> float:
    """ Returns the polygon area (perimeter) for 2D set of (x, y) points

    Args:
        polygon: (n, 2) array of xy locations for a closed polygon

    Returns:
        Area of the closed polygon
    """
    import warnings
    warnings.warn("WARNING: TESTS FOR GET_POLYGON_AREA NOT CURRENTLY PASSING USE WITH CAUTION")

    if polygon.shape[0] < 3:
        return 0
    # Due to being in 2D area refers to perimeter and volume to area
    return scipy.spatial.ConvexHull(polygon).volume


def get_point_along_line(p1: np.array, p2: np.array, percentage: float) -> np.array:
    """ Get a point n% between two points

    Args:
        p1: Line start point
        p2: Line end point
        percentage: Line percentile point, 0 being start point and 1 being end point

    Returns:
        Returns point n% along line p1->p2 where 0==p1, 0.5==midpoint of p1 and p2 and 1==p2
    """
    return p1 + ((p2 - p1) * percentage)



import math
from collections import OrderedDict
from copy import copy

import numpy as np
import skimage.morphology
import scipy.integrate
from typing import Union, Tuple

from raytils.display.utils import write_text_cv2
from raytils.perception.polygon import get_polygon_area, scale_polygon, sortpts_clockwise

__all__ = ["get_circle_blob", "get_mask_contour", "get_orientation_from_blob", "get_square_blob",
           "get_volume_from_blob"]

CITATION_FIRST_RUN_MESSAGE = False


def get_oval_blob(width: int = 100, height=50) -> np.array:
    """ Returns (max(width, height), max(width, height)) np.array filled with 0s and an oval of width*height

    Args:
        width: Width of the oval blob required
        height: Height of the oval blob required

    Returns:
        Numpy array with a oval around its center filled with 1s
    """
    canvas_size = max(width, height)
    mask, _ = get_circle_blob(canvas_size)
    mask = np.asarray([[mask[int(len(mask) * r / height)][int(len(mask[0]) * c / width)]
                        for c in range(width)] for r in range(height)])
    contour = get_mask_contour(mask)
    return mask, contour


def get_circle_blob(size: int = 100) -> np.array:
    """ Returns (size, size) np.array filled with 0s and a circle of radius=size of 1s

    Args:
        size: Diameter of the circle blob required

    Returns:
        Numpy array with a circle around its center filled with 1s
    """
    radius = size / 2
    diameter = radius * 2
    mask_size = (int(diameter), int(diameter))
    mid_point = np.array([[radius], [radius]])

    mask = np.zeros(mask_size, dtype=np.uint8)
    coords = np.asarray(np.where(mask == 0))
    distances = np.sum((coords - mid_point) ** 2, axis=0) ** 0.5
    mask[(distances <= radius).reshape(mask_size)] = 1

    # contour = np.asarray(list(zip(*np.where((np.abs(distances - radius) <= 1).reshape(mask_size)))))
    contour = get_mask_contour(mask)

    return mask, contour


def get_strawberry_blob(height=100):
    """ Returns (height, height) np.array filled with 0s and 1s of a rough strawberry blob shape

    Args:
        height: Height of the berry blob required

    Returns:
        Numpy array with a oval around its center filled with 1s
    """
    mask = np.zeros((height, height), dtype=np.uint8)
    rows = height
    mid = int(height / 2)
    min_width_top = 0.35 * rows
    max_width_top = 0.85 * rows
    top_step = (max_width_top - min_width_top) / (rows * 0.35)
    max_width_bottom = 0.85 * rows
    min_width_bottom = 0.2 * rows
    bottom_step = (max_width_bottom - min_width_bottom) / (rows * 0.65)

    for i in range(rows):
        ip = i / rows
        if ip <= 0.35:
            t = min_width_top + (top_step * i)
            t2 = t / 2
            ts = int(mid - t2)
            te = int(mid + t2)
            mask[i, ts:te] = 1.0
        else:
            t = min_width_bottom + (bottom_step * (rows - i))
            t2 = t / 2
            ts = int(mid - t2)
            te = int(mid + t2)
            mask[i, ts:te] = 1.0

    contour = get_mask_contour(mask)

    return mask, contour


def get_square_blob(size: int = 500, padding: int = 25) -> np.array:
    """ Returns (size+padding*2, size+padding*2) np.array filled with 0s and a square of 1s of size

    Args:
        size: Required square size of the blob
        padding: Required padding around the edge of the blob

    Returns:
        A Numpy array (size+padding*2, size+padding*2) filled with a square blob of 1.0s
    """
    mask_size = (int(size + padding * 2), int(size + padding * 2))
    # mid_point = np.array([[int(size + padding * 2) / 2], [int(size + padding * 2) / 2]])
    mask = np.zeros(mask_size, dtype=np.uint8)
    mask[padding:size + padding, padding:size + padding] = 1

    contour = get_mask_contour(mask)

    return mask, contour


def get_mask_contour(mask: np.array, method: str = "skimage") -> np.array:
    """ Takes a binary image (N, M) and returns the edge of the blobs in XY format

    Args:
        mask: (N, M) binary image with 1.0 indicating object pixel and 0.0 background pixel
        method: 'erosion' the method to extract the contour locations

    Returns:
        Numpy array of size (contour_length, 2) where contour length is the number of x, y points
    """
    if method not in ["erosion", "skimage"]:
        raise NotImplementedError("Method '{}' not implemented".format(method))
    if method == "skimage":
        contours = skimage.measure.find_contours(mask, 0.5)
        cnt = contours[0].astype(np.int)
        return cnt
    pad = 1
    width, height = mask.shape
    padded_mask = np.zeros((width + (pad * 2), height + (pad * 2)))
    padded_mask[pad:width + pad, pad:height + pad] = mask != 0
    return np.asarray(list(zip(*np.where(padded_mask - skimage.morphology.binary_erosion(padded_mask))))) - pad


def raw_moment(mask: np.array, i_order: int, j_order: int) -> float:
    """Get moments from mask

    Args:
        mask: (N, M) binary image with 1.0 indicating object pixel and 0.0 background pixel
        i_order: ith moment to return
        j_order: jth moment to return

    Returns:
        Returns the moment[i, j] for the given mask
    """
    nrows, ncols = mask.shape
    y_indices, x_indicies = np.mgrid[:nrows, :ncols]
    return (mask * x_indicies ** i_order * y_indices ** j_order).sum()


def get_orientation_from_blob(mask: np.array, method: str = "moments") -> Tuple[float, float, float, float]:
    """Detect Orientation From Blob

    https://alyssaq.github.io/2015/computing-the-axes-or-orientation-of-a-blob/

    Calculates orientation from the covariance matrix C:

    .. math::

        C = \\left[\\begin{array}{cc} variance(x, x) & variance(x, y)\\\\ variance(x, y) & variance(y, y) \\end{array}\\right]

    Args:
        mask: (N, M) binary image with 1.0 indicating object pixel and 0.0 background pixel
        method: 'moments' or 'eigen' decides how the covariance matrix for x, y will be calculated

    Returns:
        Tuple of x_v1, y_v1, x_v2, y_v2 (Eigen Vector of minor and major axis)
    """
    if method == "moments":
        mask_sum = mask.sum()
        m10 = raw_moment(mask, 1, 0)
        m01 = raw_moment(mask, 0, 1)
        x_centroid = m10 / mask_sum
        y_centroid = m01 / mask_sum
        u11 = (raw_moment(mask, 1, 1) - x_centroid * m01) / mask_sum
        u20 = (raw_moment(mask, 2, 0) - x_centroid * m10) / mask_sum
        u02 = (raw_moment(mask, 0, 2) - y_centroid * m01) / mask_sum
        cov = np.array([[u20, u11], [u11, u02]])
    elif method == "eigen":
        y, x = np.nonzero(mask)
        x = x - np.mean(x)
        y = y - np.mean(y)
        coords = np.vstack([x, y])
        cov = np.cov(coords)
    else:
        raise NotImplementedError("Method '{}' not yet implemented")

    eigen_values, eigen_vectors = np.linalg.eig(cov)
    sort_indices = np.argsort(eigen_values)[::-1]
    x_v1, y_v1 = eigen_vectors[:, sort_indices[0]]  # Eigenvector with the largest eigenvalue
    x_v2, y_v2 = eigen_vectors[:, sort_indices[1]]

    return x_v1, y_v1, x_v2, y_v2


def get_volume_from_blob(mask: np.array, method: str = "surface_area", limit_method: str = "contour",
        integration_method: str = "quad", n_slices: int = 100, return_slices: bool = False, radius=None, n_radius=None,
        depth=None, intrinsics=None) \
        -> Union[float, Tuple[float, list]]:
    """ Calculates volume from a 2D blob plane

    Volume comes from integration of the surface area through minor axis height

    https://doi.org/10.31256/Uk4Td6I

    Args:
        intrinsics:
        mask: (N, M) binary image with 1.0 indicating object pixel and 0.0 background pixel
        method: str 'surface_area' or 'ellipsoid'
        limit_method: 'width', 'radius', or 'contour' decides how the radius parameter is calculated
        integration_method: 'trapz' (trapezoidal rule) or 'simps' (simpsons rule)
        n_slices: The number of steps to integrate over (larger more accurate, lower faster computation)
        return_slices: If true the contours of the integral slices are returned
        radius: If limit method is radius this parameter must not be None
        n_radius: If limit method is radius and method is ellipsoid this parameter used for ellipsoidal computation

    Returns:
        Volume of the 2D binary blob in mask or (Volume, Contour Slice Locations) if return_slices is true
    """
    global CITATION_FIRST_RUN_MESSAGE
    if not CITATION_FIRST_RUN_MESSAGE:
        print("Please cite https://doi.org/10.31256/Uk4Td6I if using this work")
        CITATION_FIRST_RUN_MESSAGE = True
    contour = get_mask_contour(mask, method="erosion")

    # How to calculate the radius parameter for integration
    if limit_method == "width":  # Take the width of the mask image
        width = mask.shape[0]
        height = mask.shape[1]
        n_radius = height / 2
        radius = width / 2
    elif limit_method == "height":  # Take the width of the mask image
        width = mask.shape[0]
        height = mask.shape[1]
        n_radius = width / 2
        radius = height / 2
    elif limit_method == "contour":  # Contour take the widest section of the x and y contours
        radius = (np.ptp(contour[:, 0])) / 2
        height = (np.ptp(contour[:, 1]))
        n_radius = height / 2
    elif limit_method == "radius":
        if radius is None:
            raise ValueError("Cannot use method radius when parameter radius=None")
        radius = radius
    else:
        raise NotImplementedError("Method '{}' not yet implemented".format(limit_method))

    contour = get_mask_contour(mask, method="erosion")

    my, mx = np.mean(np.where(mask != 0), axis=1)
    center_of_mass = np.asarray([[mx, my]])

    if intrinsics is not None:
        fx, fy, cx, cy = intrinsics

        depth[np.logical_not(mask)] = 0

        y, x = np.where(mask != 0)
        z = depth[np.logical_and(depth != 0, mask != 0)]
        z_max = np.max(z)
        x = (x - cx) * z_max / fx
        y = (y - cy) * z_max / fy
        minor = np.ptp(x)
        major = np.ptp(y)
        cross = np.ptp(z) * 2

        contour = contour.astype(np.float)
        contour[:, 0] = np.abs((contour[:, 0] - cy) * z_max / fy)
        contour[:, 1] = np.abs((contour[:, 1] - cx) * z_max / fx)

        center_of_mass = np.abs(np.asarray([[np.mean(x), np.mean(y)]]))
        radius = minor / 2
        n_radius = major / 2
        height = major
        # if integration_method == "quad":
        #     print(f"Minor = {minor:.2f}, Major = {major:.2f}, Cross = {cross:.2f}")

    if method == "disc":
        if intrinsics is not None:
            fa = 1
            px2x = lambda _x: (_x - cx) * z_max / fx
            py2y = lambda _y: (_y - cy) * z_max / fy
        else:
            fa = 0
            px2x = lambda _x: _x
            py2y = lambda _y: _y
        binary = mask != 0
        row_sums = binary.sum(axis=fa)
        row_start = (binary != 0).argmax(axis=fa)
        sample_rate = 1

        row_start = row_start[row_sums != 0]
        row_sums = row_sums[row_sums != 0]
        idxs = range(0, len(row_start), sample_rate)
        row_start = row_start[idxs]
        row_sums = row_sums[idxs]

        row_sum_starts = row_start
        row_sum_ends = row_start + row_sums
        rows_radius = (px2x(row_sum_ends) - px2x(row_sum_starts)) / 2.0
        dy_distance = (py2y(sample_rate) - py2y(0))

        disc_areas = (np.pi * (rows_radius ** 2))
        disc_volumes = (np.pi * (rows_radius ** 2) * dy_distance)

        polygon_volume = disc_volumes.sum()

        if return_slices:
            contours = [contour]
            return polygon_volume, contours
        return polygon_volume
    elif method == "surface_area":
        # Derivative of Area is Perimeter, Derivative of Volume is Surface Area
        # http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.598.3691&rep=rep1&type=pdf
        # Setup the integration arrays (x=radius at each slice, y=area function at each slice (f(x)))
        # TODO: Use sets to represent the contours so unique points aren't duplicated
        contours = []
        dy = radius / n_slices
        radius_steps = np.arange(0, radius, dy)  # x=r at each step
        actual_steps = []  # Array for quad integration step tracking

        # y = f(x), where f(x) is the surface area at a radius x
        def f(_x):
            actual_steps.append(_x)
            p = scale_polygon(contour, center_of_mass, _x / radius)
            a = get_polygon_area(p)
            contours.append(p)
            return 4 * a

        # Summation to solve for volume
        if integration_method == "quad":
            polygon_volume, polygon_error = scipy.integrate.quad(f, 0, radius)
            radius_steps = actual_steps
        elif integration_method == "trapz":
            y = [f(x) for x in radius_steps]
            polygon_volume = scipy.integrate.trapz(y, radius_steps)
        elif integration_method == "simps":
            y = [f(x) for x in radius_steps]
            polygon_volume = scipy.integrate.simps(y, radius_steps)
        else:
            raise NotImplementedError("Integration method '{}' is not yet implemented".format(integration_method))

        if return_slices:
            return polygon_volume, contours
        return polygon_volume
    elif method == "ellipsoid":
        polygon_volume = (4 / 3) * np.pi * n_radius * radius ** 2
        if n_radius is None:
            raise ValueError("You must pass n_radius")
        if return_slices:
            contours = [contour]
            return polygon_volume, contours
        return polygon_volume
    else:
        raise NotImplementedError("Method '{}' not yet implemented".format(method))



#  Raymond Kirk (Tunstill) Copyright (c) 2020
#  Email: ray.tunstill@gmail.com


class _unknown_class:
    def __init__(self):
        pass

    def __getitem__(self, item):
        return "class {}".format(item)


class ModelFromConfig:
    def __init__(self, config_file, model_file=None):
        from detectron2.config import get_cfg
        from detectron2.data import MetadataCatalog, DatasetCatalog
        from detectron2.engine.defaults import DefaultPredictor

        self.cfg = get_cfg()
        self.classes = _unknown_class()

        try:
            from fruit_detection.config import add_fruit_detection_config
            from fruit_detection.datasets import register_data_set
            add_fruit_detection_config(self.cfg)
            self.cfg.merge_from_file(config_file)
            register_data_set(self.cfg.DATASETS.TRAIN[0])
            dataset = DatasetCatalog.get(self.cfg.DATASETS.TRAIN[0])
            metadata = MetadataCatalog.get(self.cfg.DATASETS.TRAIN[0])
            self.classes = metadata.get("thing_classes", None)
            if not self.classes:
                self.classes = _unknown_class
        except ImportError:
            self.cfg.merge_from_file(config_file)

        if model_file is not None:
            self.cfg.MODEL.WEIGHTS = model_file
        self.cfg.freeze()

        self.predictor = DefaultPredictor(self.cfg)

    def __call__(self, image):
        return self.predictor(image)

    @property
    def model(self):
        return self.predictor.model



from collections.abc import Hashable
from ast import literal_eval
from typing import Dict, Union, Any, Iterable, Sized, Type

from raytils.ui.formatting import suffix_join

try:
    input = raw_input
except NameError:
    pass


def get_string_type(input_data: str) -> Type:
    """Returns the type of a string by evaluating it"""
    try:
        return type(literal_eval(input_data))
    except (ValueError, SyntaxError):
        return str


def get_input(prompt: str = "", true_type: bool = False) -> Any:
    """Returns input from the user and optionally converts it to the appropriate python type"""
    response = input(prompt)

    if true_type:
        string_type = get_string_type(response)
        return string_type(response)

    return response


def get_selection_from_dict(options: Dict[Union[str, int, float], Any], required: bool = True) -> Any:
    """Prompts the user for a choice from a given selection of dict values using their keys as the selectors

    Args:
        options: A dictionary of options (values) and selectors (keys)
        required: If true the function will not return until a valid selection is made, if false it can return None

    Returns:
        A value from the options dictionary or possibly None if required is False
    """
    prompt = "Enter Selection: "
    valid_selection = False
    selection = 0

    options_string = "Options: \n\t" + '\n\t'.join(["{}: {}".format(k, v) for k, v in options.items()])
    print(options_string)

    while not valid_selection:
        selection = get_input(prompt, true_type=True)
        if not isinstance(selection, Hashable) or selection not in options:
            if not required:
                return None
            valid_options = suffix_join((["{} ({})".format(k, str(type(k))) for k in options.keys()]), ', ', ' or ')
            print("'{}' ({}) is not a valid option, please choose from: {}".format(selection, type(selection),
                                                                                   valid_options))
            continue
        valid_selection = True

    return options[selection]


def get_selection_from_list(options: list, required: bool = True) -> Any:
    """Prompts the user for a choice from a given selection of list values using integers as selection options

    Args:
        options: A list of options for selection
        required: If true the function will not return until a valid selection is made, if false it can return None

    Returns:
        A value from the options list or possibly None if required is False
    """
    return get_selection_from_dict({i: v for i, v in enumerate(options)}, required)


def __test_selections():
    test_options_list = ["Selection Zero", "Two"]
    test_options_dict = {1.2: 1.2, "Zero": "Selection Zero", 2: "Two"}
    selection = get_selection_from_dict(test_options_dict)
    print("Selected:", selection)
    selection = get_selection_from_list(test_options_list)
    print("Selected:", selection)
    selection = get_selection_from_list(test_options_list, required=False)
    print("Selected:", selection)



if __name__ == '__main__':
    __test_selections()




def get_time_difference(time1, time2, accurate=False):
    units = {1: ("millisecond", 0.01), 60: ("second", 1), 3600: ("minute", 60), 86400: ("hour", 3600),
             float("inf"): ("day", 86400)}

    seconds_difference = time1 - time2
    type_conversion = float if accurate else int
    format_string = "{}".format("{:.1f} {}" if accurate else "{} {}")

    for unit_threshold, (unit_string, seconds_to_unit) in units.items():
        if seconds_difference <= unit_threshold:
            unit_difference = type_conversion(seconds_difference / seconds_to_unit)
            return format_string.format(unit_difference, unit_string) + ("" if unit_difference <= 1 else "s")

    raise ValueError(f"Input time is invalid '{time1}' and '{time2}', cannot parse.")


def get_abs_time_difference(time1, time2, accurate=True):
    return get_time_difference(max(time1, time2), min(time1, time2), accurate)



from threading import Thread, Event
from typing import Callable, Optional, Union

import cv2
import numpy as np

from raytils.time import FPSCounter

__all__ = ["OpenCVDisplay"]


class OpenCVDisplay:
    """ Easy class for creating a GUI on another thread but still keeping the waitKey and interaction functionality

    Attributes:
        window_name: Name of the window to display
        update_rate: Number of times to update the display per second
        key_press_callback: Callback to call when on key press, the callback takes the key code as it's first argument
    """

    def __init__(self, window_name: str = "", update_rate: int = 0, key_press_callback: Callable = None):
        self.window_name = window_name
        update_rate = 1000 if update_rate <= 0 else update_rate
        self.__update_delay = int((1.0 / update_rate) * 1000)
        self.__key_press_callback = key_press_callback
        self.__key_press_callbacks = {}
        self.__last_key_pressed = -1

        self.__thread = None
        self.__frame = None
        self.__is_shutdown = Event()
        self.__is_shutdown.set()
        self.__key_pressed = Event()
        self.__display_fps_counter = FPSCounter(queue_length=update_rate * 2)
        self.start()

    def start(self):
        """Starts the display and returns the class reference

        It's automatically called on the construction of the object, it's only necessary to recall it, if stop has been
        called.
        """
        if self.is_running():
            return self
        self.__is_shutdown.clear()
        self.__thread = Thread(target=self.__update, args=())
        self.__thread.start()
        return self

    def is_running(self):
        """Returns true if the display hasn't been closed (stop() called)"""
        return not self.__is_shutdown.is_set()

    def __update(self):
        cv2.namedWindow(self.window_name, cv2.WINDOW_GUI_EXPANDED)
        cv2.destroyAllWindows()

        while self.is_running():
            if self.__frame is None:
                continue

            self.__display_fps_counter.tic()
            cv2.imshow(self.window_name, self.__frame)
            self.__last_key_pressed = cv2.waitKey(self.__update_delay)
            self.__key_pressed.set()
            self.__key_pressed.clear()
            if self.__last_key_pressed != -1:
                self.send_key(self.__last_key_pressed)
            self.__display_fps_counter.toc()

    def show(self, frame: np.array, get_key: bool = False) -> Optional[int]:
        """Updated the display with a new frame and optionally returns the key pressed on the frame update cycle

        Args:
            frame: Numpy image
            get_key: If true the function will return the key pressed similar to cv2.waitKey

        Returns:
            The key pressed if get_key is true
        """
        self.__frame = frame
        if self.is_running() and get_key:
            return self.get_key()

    def get_key(self) -> int:
        """Waits for and returns the next key press"""
        self.__key_pressed.wait()
        return self.__last_key_pressed

    def send_key(self, key: Union[str, int]):
        """Simulates the key press event on the display

        Used internally but useful for tests

        Args:
            key: The char or key code representation of a key you would like to simulate pressing.

        Raises:
            TypeError: If the key code is malformed
        """
        if isinstance(key, str):
            key = ord(key)

        if callable(self.__key_press_callback):
            Thread(target=self.__key_press_callback, args=(key,)).start()
        if key in self.__key_press_callbacks:
            callback, args = self.__key_press_callbacks[key]
            Thread(target=callback, args=args).start()

    def stop(self):
        """Closes and stops the display thread"""
        cv2.destroyAllWindows()
        self.__is_shutdown.set()
        if self.__thread.is_alive():
            self.__thread.join()

    def get_fps(self) -> float:
        """Returns the fps (frames per second) of the display update cycle"""
        return self.__display_fps_counter.get_fps()

    def register_key(self, key: Union[str, int], callback: Callable, args: Union[tuple, list] = ()):
        """ Register a function to a key press action

        Example:

        .. code-block:: python

            display = OpenCVDisplay()
            # Close the display if q or escape (127) is pressed
            display.register_key('q', callback=display.stop)
            display.register_key(127, callback=display.stop)

        Args:
            key: The char or key code representation of a key you would like to bind a function to.
            callback: The function you would like to bind to the key press.
            args: Extra arguments you would like to pass to the callback function

        Raises:
            TypeError: If the key code is malformed, or the function is not callable
        """
        if not callable(callback):
            raise TypeError("The callback parameter must be callable")
        if not isinstance(key, (str, int)):
            raise TypeError("The key must be an integer key code, or a string of length 1")

        if isinstance(key, str):
            key = ord(key)

        self.__key_press_callbacks[key] = (callback, tuple(args))

    def unregister_key(self, key: Union[str, int]):
        """Unregister the callback for a specified key

        Args:
            key: The char or key code representation of a key you would like to unbind a function from.

        Raises:
            TypeError: If key is an invalid key code
        """
        if isinstance(key, str):
            key = ord(key)
        del self.__key_press_callbacks[key]

    def unregister_all_keys(self):
        """Unregister the callback for all key presses"""
        self.__key_press_callbacks = {}

    def __exit__(self, exc_type, exc_value, traceback):
        cv2.destroyAllWindows()



from collections import deque
from timeit import default_timer as timer

__all__ = ["FPSCounter"]


class FPSCounter:
    """Class for calculating the FPS of an operation

    Usage:

    .. code-block:: python

        import time
        fps_counter = FPSCounter()

        while True:
            fps_counter.tic()
            time.sleep(0.2) # Operations you want to time
            fps_counter.toc()
            print("FPS", fps_counter.get_fps())

    Attributes:
        queue_length: Number of times to average over for the get_fps function
    """
    NOT_INITIALIZED_VALUE = -1

    def __init__(self, queue_length: int = 30):
        self.__frame_times = deque(maxlen=int(queue_length))
        self.__start = None

    def reset(self):
        """Clears the timer history"""
        self.__frame_times.clear()

    def start(self):
        """Starts the timer"""
        self.__start = timer()

    def tic(self):
        """Starts the timer"""
        self.start()

    def stop(self):
        """Ends the timer and adds elapsed time since start() or tic() called to history"""
        if not self.__start:
            return
        self.__frame_times.append(timer() - self.__start)
        self.__start = None

    def toc(self):
        """Ends the timer and adds elapsed time since start() or tic() called to history"""
        self.stop()

    def get_fps(self) -> float:
        """Returns the FPS"""
        time_sum = sum(self.__frame_times)
        return (len(self.__frame_times) / time_sum) if time_sum else self.NOT_INITIALIZED_VALUE

    def get_times(self):
        """Returns the total time, ms and FPS"""
        if len(self.__frame_times) == 0:
            return self.NOT_INITIALIZED_VALUE, self.NOT_INITIALIZED_VALUE, self.NOT_INITIALIZED_VALUE
        total = sum(self.__frame_times)
        time_avg = sum(self.__frame_times) / len(self.__frame_times)
        ms_avg, fps = time_avg * 1000, 1. / time_avg
        return total, ms_avg, fps

    def __str__(self):
        return "Total Time: {:.2f}s Avg Time: {:.2f}ms ({:.2f} fps)".format(*self.get_times())



from PIL import Image, ExifTags


def get_image_size(filepath):
    """Return image size (width, height) without loading. Respects correct orientation."""
    image = Image.open(filepath)

    try:
        orientation = None
        for orientation in ExifTags.TAGS.keys():
            if ExifTags.TAGS[orientation] == 'Orientation':
                break

        exif = dict(image.getexif().items())

        if exif[orientation] == 3:
            image = image.transpose(Image.ROTATE_180)
        elif exif[orientation] == 6:
            image = image.transpose(Image.ROTATE_270)
        elif exif[orientation] == 8:
            image = image.transpose(Image.ROTATE_90)
        image.close()
    except (AttributeError, KeyError, IndexError):
        pass  # Image doesn't have ExIF data

    return image.size



# Create class for creating a regression report later (inspired by sklearn/classification_report)
from collections import Callable

import numpy as np
import pandas as pd
from sklearn.metrics import (
    explained_variance_score,
    max_error,
    mean_absolute_error,
    mean_squared_error,
    mean_squared_log_error,
    median_absolute_error,
    r2_score,
    mean_poisson_deviance,
    mean_gamma_deviance,
    mean_tweedie_deviance,
)
import plotly.express as px
from io import StringIO


class Metrics:
    def __init__(self, drop_zeros=False):
        """A collection of regression metrics
        Args:
            drop_zeros: If true drop any zeros in y_true and the corresponding value in y_pred for percentage based
             metrics
        """
        self._drop_zeros = drop_zeros

    def calculate_all(self, y_true, y_pred):
        _computed_metrics = {}
        valid_class_functions = [
            x for x in dir(self) if x != "calculate_all" and not x.startswith("_") and "__" not in x and
            isinstance(getattr(self, x), Callable)
        ]
        for class_attribute in valid_class_functions:
            metric_name, metric_value = class_attribute, float("NaN")
            try:
                metric_value = getattr(self, class_attribute)(y_true, y_pred)
            except Exception:
                pass
            _computed_metrics[metric_name] = metric_value
        return _computed_metrics

    def residual_sum_of_squares(self, y_true, y_pred):
        rss = np.sum(np.square(y_true - y_pred))
        return rss

    def mean_squared_error(self, y_true, y_pred):
        mse = mean_squared_error(y_true, y_pred)
        return mse

    def mean_squared_percentage_error(self, y_true, y_pred):
        idx = np.s_[:]
        if self._drop_zeros:
            idx = y_true != 0
        y, y_hat = y_true[idx], y_pred[idx]
        mspe = np.mean(np.divide(np.square(y - y_hat), np.square(y)))
        return mspe

    def root_mean_square_error(self, y_true, y_pred):
        rmse = mean_squared_error(y_true, y_pred, squared=False)
        return rmse

    def root_mean_square_percentage_error(self, y_true, y_pred):
        idx = np.s_[:]
        if self._drop_zeros:
            idx = y_true != 0
        y, y_hat = y_true[idx], y_pred[idx]
        rmspe = np.square(np.mean(np.divide(np.square(y - y_hat), np.square(y))))
        return rmspe

    def mean_absolute_error(self, y_true, y_pred):
        mae = mean_absolute_error(y_true, y_pred)
        return mae

    def mean_absolute_percentage_error(self, y_true, y_pred):
        idx = np.s_[:]
        if self._drop_zeros:
            idx = y_true != 0
        y, y_hat = y_true[idx], y_pred[idx]
        mape = np.mean(np.abs(np.divide((y - y_hat), y)))
        return mape

    def relative_squared_error(self, y_true, y_pred):
        rse = np.divide(np.sum(np.square(y_true - y_pred)),
                        np.sum(np.square(y_true - np.mean(y_true))))
        return rse

    def relative_absolute_error(self, y_true, y_pred):
        rae = np.divide(np.sum(np.abs(y_true - y_pred)),
                        np.sum(np.abs(y_true - np.mean(y_true))))
        return rae

    def r2_score(self, y_true, y_pred):
        # Coefficient of determination
        r2 = r2_score(y_true, y_pred)
        return r2

    def explained_variance_score(self, y_true, y_pred):
        return explained_variance_score(y_true, y_pred)

    def max_error(self, y_true, y_pred):
        return max_error(y_true, y_pred)

    def min_error(self, y_true, y_pred):
        return np.min(np.abs(y_true - y_pred))

    def mean_squared_log_error(self, y_true, y_pred):
        return mean_squared_log_error(y_true, y_pred)

    def median_absolute_error(self, y_true, y_pred):
        return median_absolute_error(y_true, y_pred)

    def mean_poisson_deviance(self, y_true, y_pred):
        return mean_poisson_deviance(y_true, y_pred)

    def mean_gamma_deviance(self, y_true, y_pred):
        return mean_gamma_deviance(y_true, y_pred)

    def mean_tweedie_deviance(self, y_true, y_pred):
        return mean_tweedie_deviance(y_true, y_pred)


class RegressionReport:
    """Generate regression report for two arrays similar to sklearn classification report

    Examples:
        To run a report, plot and get specific metrics

        >>> report = RegressionReport(np.array([0.5, 1.0]), np.array([1.0, 0.5]))  # generate report for two arrays
        >>> fig = report.plot()  # plot using plotly express
        >>> report.print()  # print(report)
        Regression Report:
                 explained_variance_score:               -3.000
                                max_error:                0.500
                      mean_absolute_error:                0.500
           mean_absolute_percentage_error:                0.750
                      mean_gamma_deviance:                0.500
                    mean_poisson_deviance:                0.347
                       mean_squared_error:                0.250
                   mean_squared_log_error:                0.083
            mean_squared_percentage_error:                0.625
                    mean_tweedie_deviance:                0.250
                    median_absolute_error:                0.500
                                min_error:                0.500
                                 r2_score:               -3.000
                  relative_absolute_error:                2.000
                   relative_squared_error:                4.000
                  residual_sum_of_squares:                0.500
                   root_mean_square_error:                0.500
        root_mean_square_percentage_error:                0.391
        Percentiles:
                                        5:               -0.450
                                       25:               -0.250
                                       50:                0.000
                                       75:                0.250
                                       95:                0.450
        >>> r2 = report["r2_score"]  # get specific metric
        -3.000
    """

    def __init__(self, y_true, y_pred, drop_zeros=False, baselines=False):
        """

        Args:
            y_true: 1d np.array of actual y values example=np.array([1,2,3,4,5])
            y_pred: 1d np.array of predicted y values example=np.array([1,2,3,4,5])
            drop_zeros: If true drop any zeros in y_true and the corresponding value in y_pred for percentage based
             metrics
            baselines: If True RegressionReport will include the metrics using predictions as the last known value. If
             int then will use last (+ve int) or future known (-ve int) to calculate baseline achievable accuracy.
        """
        self._y_true, self._y_pred = self._check_arrays(y_true, y_pred)

        # Some forecasting models drop zeros https://ideas.repec.org/a/for/ijafaa/y2006i4p32-35.html
        metrics = Metrics(drop_zeros=drop_zeros)

        self._computed_metrics = metrics.calculate_all(self._y_true, self._y_pred)
        self._baselines = bool(baselines)
        self._baseline_metrics = None
        if self._baselines:
            shift_by = int(baselines)
            self._baseline_pred = self._shift_array(self._y_true, shift_by=shift_by, fill_value=None)
            self._baseline_metrics = metrics.calculate_all(self._y_true, self._baseline_pred)

    @staticmethod
    def _check_arrays(y_true, y_pred):
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        return y_true, y_pred

    def __str__(self):
        report_str = []
        fmt_string = "{:>" + str(max([len(x) for x in self._computed_metrics.keys()])) + "s}: {: >20.3f}"
        if self._baselines:
            fmt_string += " ({:.3f})"
        report_str.append('Regression Report:')
        error_metrics = []

        for metric_name, metric_value in self._computed_metrics.items():
            if self._baselines:
                report_str.append(fmt_string.format(metric_name, metric_value, self._baseline_metrics[metric_name]))
            else:
                report_str.append(fmt_string.format(metric_name, metric_value))
            if metric_value == float("NaN"):
                error_metrics.append(metric_name)

        report_str.append('\nPercentiles:')
        percentile = [5, 25, 50, 75, 95]
        percentile_value = np.percentile(self._y_true - self._y_pred, percentile)
        baseline_percentile_value = np.empty_like(percentile_value)
        if self._baselines:
            baseline_percentile_value = np.percentile(self._y_true - self._baseline_pred, percentile)
        for p, pv, blv in zip(percentile, percentile_value, baseline_percentile_value):
            if self._baselines:
                report_str.append(fmt_string.format(str(p), pv, blv))
            else:
                report_str.append(fmt_string.format(str(p), pv))
        if len(error_metrics):
            report_str.append("\nCould not compute: {}".format(len(error_metrics), ", ".join(error_metrics)))
        return '\n'.join(report_str)

    def __repr__(self):
        return self.__str__()

    def print(self):
        print(self)

    def __getitem__(self, item):
        return self._computed_metrics[item]

    def plot(self, target_name=None, show=True):
        labels = ["y (Truth)", "ŷ (Prediction)"]
        title = "Ground Truth vs Predictions"
        yaxis_title = ""
        if target_name:
            target_name = target_name + (" " * (not target_name.endswith(" ")))
            labels = [target_name + x for x in labels]
            title = target_name + title
            yaxis_title = target_name
        labels = ["Time Index"] + labels
        data_array = [np.arange(len(self._y_true)), self._y_true, self._y_pred]
        if self._baselines:
            labels.append("y (Baseline)")
            data_array.append(self._baseline_pred)
        df = pd.DataFrame(dict(zip(labels, data_array)))
        fig = px.line(df, x=labels[0], y=labels[1:], title=title)
        fig.update_layout(yaxis_title=yaxis_title)
        if show:
            fig.show()
        return fig

    @staticmethod
    def _shift_array(data, shift_by, fill_value=None):
        """Shift array by n values, fill with first known value if fill_value is None"""
        result = np.empty_like(data)
        if shift_by > 0:
            if fill_value is None:
                fill_value = data[:-shift_by][0]
            result[:shift_by] = fill_value
            result[shift_by:] = data[:-shift_by]
        elif shift_by < 0:
            if fill_value is None:
                fill_value = data[-shift_by:][0]
            result[shift_by:] = fill_value
            result[:shift_by] = data[-shift_by:]
        else:
            result[:] = data
        return result



__all__ = ["print_variables"]


def print_variables(sep=', ', association=" = ", *args, **kwargs):
    r""" Easy function to visualise variables in development

    Usage:

    .. code-block:: python

        x = 5
        y = 10
        string_pos = "({}, {})".format(x, y)
        print_variables(x=x, y=y, location=string_pos) # Output -> 'X = 5, Y = 10, Location = (5, 10)'

    Args:
        sep: How to separate different parameters e.g ", " -> "v1=1, v2=2" or "\\n" -> r"v1=1\\nv2=2"
        association: The link between variable names and values, default is equality (" = ") e.g v1_name = v1_value
        *args: Optional Arguments to pass to the print function
        **kwargs: The keyword value list passed to the print variable function

    """
    print_string = sep.join(["{}{}{}".format(
        ' '.join(map(lambda x: x[0].upper() + x[1:].lower(), k.split('_'))), association, v)
        for k, v in kwargs.items()])
    print(print_string, *args)



import numpy as np
import cv2

__all__ = ["image_list_to_grid", "write_text_cv2"]


def write_text_cv2(image, text, x=None, y=None, centre=True, font=None, scale=None, thickness=None, colour=None):
    """Place text on an image around a point (x, y) top left (centre=False) or centre (centre=True)

    Args:
        image: np.array to draw on (m, n, 3)
        text: Text to be put on the image
        x: text co-ordinate x
        y: text co-ordinate y
        centre: If true (x, y) anchor is the centre not top-left
        font: cv2.FONT (default = cv2.FONT_HERSHEY_SIMPLEX)
        scale: Text scale (default = 1)
        thickness: Text thickness (default = 1)
        colour: Colour to draw text with (default = (0, 0, 0)

    Returns:
        image: With text drawn
    """
    if y is None:
        y = image.shape[0] / 2
    if x is None:
        x = image.shape[1] / 2

    x = image.shape[1] // 2 if x is None else x
    y = image.shape[0] // 2 if y is None else y
    font = cv2.FONT_HERSHEY_SIMPLEX if font is None else font
    scale = 1 if scale is None else scale
    thickness = 1 if thickness is None else thickness
    colour = (0, 0, 0) if colour is None else colour
    text = str(text)

    text_size = cv2.getTextSize(text, font, scale, thickness)[0]

    if centre:
        x -= text_size[0] / 2
        y -= text_size[1] / 2

    cv2.putText(image, text, (int(x), int(y)), font, scale, colour, thickness)
    return image


def image_list_to_grid(images_list: list, columns: int = None, resize: bool = False, background_colour=None):
    """Convert a list of Numpy nd.array images to a single grid image

    Args:
        images_list: List of numpy images
        columns: Desired number of columns (optional)
        resize: If the images are of different sizes this will  (optional)

    Returns:
        Grid of all images in the image_list
    """
    max_size = images_list[0].shape
    for image in images_list:
        max_size = tuple(max(max_size[i], image.shape[i]) for i in range(len(max_size)))

    if background_colour is None:
        background_colour = (0, 0, 0)

    center_h = max_size[0] / 2
    center_w = max_size[1] / 2
    canvas = []
    for image in images_list:
        if resize:
            canvas.append(cv2.resize(image, max_size[:2]))
        else:
            c = np.ones(max_size, dtype=image.dtype) * background_colour
            ch, cw = image.shape[:2]
            ch, cw = ch / 2, cw / 2
            sh, eh, sw, ew = int(center_h - ch), int(center_h + ch), int(center_w - cw), int(center_w + cw),
            c[sh:eh, sw:ew] = image
            canvas.append(c)

    if columns is None:
        columns = int((canvas[0].shape[0] / canvas[0].shape[1]) * (len(canvas) ** 0.5)) + 1
    pad_by = (columns - (len(canvas) % columns))
    pad_by = 0 if pad_by == columns else pad_by
    padding = [np.zeros_like(canvas[-1])] * pad_by
    padded_canvas = canvas + padding

    rows = len(padded_canvas) // columns

    padded_canvas = np.asarray(padded_canvas)
    length, height, width, intensity = padded_canvas.shape

    result = padded_canvas.reshape(rows, columns, height, width, intensity).swapaxes(1, 2).reshape(
        height * rows, width * columns, intensity)

    return result



from collections import deque
from timeit import default_timer as timer
from typing import Callable

__all__ = ["FunctionTime", "function_timer"]


class FunctionTime:
    """ Decorator class for creating function timer objects

    Available functions are interval_logger and logger:

    .. code-block:: python

        import time
        function_timer = FunctionTime() # or from raytils.time import function_timer

        # Print the time every nth (5) run
        @function_timer.interval_logger(5)
        def time_me():
            time.sleep(1)

        # Print the time every run
        @function_timer.logger()
        def time_me_always():
            time.sleep(2)

        while True:
            time_me()
            time_me_always()

    Attributes:
        smooth_window: Number of past values to average the time over
        log_function: The function used to output the time (default is print)
    """

    def __init__(self, smooth_window: int = 30, log_function: Callable = None):
        self.counter = 0
        if log_function is None:
            log_function = print
        self.log_function = log_function
        self.func_times = deque(maxlen=smooth_window)

    def copy(self):
        """Returns new FunctionTime object, allows objects to easily be duplicated in one line"""
        return FunctionTime(self.func_times.maxlen, log_function=self.log_function)

    def interval_logger(self, interval: int) -> Callable:
        """ Decorator to log the function time every interval runs

        Args:
            interval: When to log the time (every interval runs)
        """
        # Always return new instance of interval logger
        return self.copy()._interval_logger(interval)

    def _interval_logger(self, interval: int) -> Callable:
        interval = interval * (0 < interval)
        self.counter = self.counter * (interval > self.counter)

        def function_time_decorator(method: Callable):
            def timed(*args, **kwargs):
                ts = timer()
                result = method(*args, **kwargs)
                self.func_times.append(timer() - ts)
                self.counter += 1
                if interval == self.counter:
                    self.log_function("{} {:.2f}ms ({:.2f} fps)".format(method.__name__, *self.__get_times()))
                    self.counter = 0
                return result

            return timed

        return function_time_decorator

    def logger(self, method: Callable) -> Callable:
        """ Decorator to log the function time every run"""
        return self.copy()._logger(method)

    def _logger(self, method: Callable) -> Callable:
        def timed(*args, **kwargs):
            ts = timer()
            result = method(*args, **kwargs)
            self.func_times.append(timer() - ts)
            self.log_function("{} {:.2f}ms ({:.2f} fps)".format(method.__name__, *self.__get_times()))
            return result

        return timed

    def __get_times(self):
        time_avg = sum(self.func_times) / len(self.func_times)
        ms, fps = time_avg * 1000, 1. / time_avg
        return ms, fps


function_timer = FunctionTime()



import numpy as np

__all__ = ["Intrinsics", "deproject_depth_image", "get_all_xy_pixel_coords", "get_all_yx_pixel_coords"]


class Intrinsics:
    """ Creates a projection matrix from fx, fy, cx, cy parameters

    Attributes:
        fx: Focal length of the x axis in pixels
        fy: Focal length of the y axis in pixels
        cx: Principal point of the x axis in pixels
        cy: Principal point of the y axis in pixels
        skew: The image skew in pixels
    """
    def __init__(self, fx: float, fy: float, cx: float, cy: float, skew: float = 0):
        self._fx = fx
        self._fy = fy
        self._cx = cx
        self._cy = cy
        self._skew = skew
        self.K = np.array([[fx, skew, cx], [0, fy, cy], [0, 0, 1]])

    @property
    def projection_matrix(self) -> np.array:
        """Returns the projection matrix K (3, 3) from [[fx, skew, cx], [0, fy, cy], [0, 0, 1]]"""
        return self.K


def get_all_xy_pixel_coords(height: int, width: int) -> np.array:
    """ Get pixel co-ordinates in x, y format (col major)

    Args:
        height: The height of the image
        width: The width of the image

    Returns:
        Numpy array of (height*width, 2) x, y coordinates for an image of size (height, width)
    """
    row_indices = np.arange(height)
    col_indices = np.arange(width)
    # Mesh grid col-row to get x, y not y, x
    pixel_grid = np.meshgrid(col_indices, row_indices)
    pixels_coords = np.c_[pixel_grid[0].flatten(), pixel_grid[1].flatten()].T
    return pixels_coords


def get_all_yx_pixel_coords(height: int, width: int) -> np.array:
    """ Get pixel co-ordinates in y, x format (row major)

    Args:
        height: The height of the image
        width: The width of the image

    Returns:
        Numpy array of (height*width, 2) y, x coordinates for an image of size (height, width)
    """
    row_indices = np.arange(height)
    col_indices = np.arange(width)
    # Mesh grid col-row to get x, y not y, x
    pixel_grid = np.meshgrid(row_indices, col_indices)
    pixels_coords = np.c_[pixel_grid[0].flatten(), pixel_grid[1].flatten()].T
    return pixels_coords


def deproject_depth_image_exp(depth_image: np.array, fx: float, fy: float, cx: float, cy: float):
    """ Deproject depth image into a point cloud image of shape [height, width, 3(x, y, z)]

    Args:
        depth_image: Numpy image of depth values in pixel format (int) of shape (height, width)
        fx: Focal length of the x axis in pixels
        fy: Focal length of the y axis in pixels
        cx: Principal point of the x axis in pixels
        cy: Principal point of the y axis in pixels

    Returns:
        Point cloud image of shape [height, width, 3(x, y, z)]
    """
    # Reformat intrinsics to 3x3 projection matrix
    intrinsics = Intrinsics(fx, fy, cx, cy)
    K = intrinsics.projection_matrix

    height, width = depth_image.shape[:2]
    pixels_coords = get_all_xy_pixel_coords(height, width)

    # Reshape pixels_coords from (height*width, 2) to (3, height*width) where [3] is a height*width array of 1.0
    pixels_homogeneous = np.r_[pixels_coords, np.ones([1, pixels_coords.shape[1]])]
    depth_array = np.tile(depth_image.data.flatten(), [3, 1])

    # Deproject to depth image
    points_3d = depth_array * np.linalg.inv(K).dot(pixels_homogeneous)
    point_cloud_image = points_3d.reshape(depth_image.height, depth_image.width, 3)
    return point_cloud_image


def deproject_depth_image(depth: np.ndarray, fx: float, fy: float, cx: float, cy: float, scale=10000.0):
    """ Deproject depth image into a point cloud image of shape [height, width, 3(x, y, z)]

    Args:
        depth: Numpy image of depth values in pixel format (int) of shape (height, width)
        fx: Focal length of the x axis in pixels
        fy: Focal length of the y axis in pixels
        cx: Principal point of the x axis in pixels
        cy: Principal point of the y axis in pixels
        scale: Scale of the point cloud (mm, cm, m etc)

    Returns:
        Points image of shape [height, width, 3(x, y, z)]
    """
    z_pos = (depth / scale).flatten()
    yx_pos = np.where(np.ones(depth.shape))

    points = np.ndarray((len(z_pos), 3))
    points[:, 0] = (yx_pos[1] - cx) * z_pos / fx
    points[:, 1] = ((yx_pos[0] - cy) * z_pos / fy) * -1
    points[:, 2] = z_pos * -1

    return points



import inspect


def suffix_join(array: list, sep: str = '', suffix_sep: str = None):
    """Returns similarly to str.join() however allows you to specify a different final join separator

    Example:
        ', '.join(["One", "Two", "Three"])  # => 'One, Two, Three'
        suffix_join(["One", "Two", "Three"], sep=', ', suffix_sep=" or ")  # => 'One, Two or Three'
    """
    if suffix_sep is None:
        suffix_sep = sep
    if len(array) >= 2:
        return sep.join(array[:-1]) + suffix_sep + array[-1]
    return sep.join(array)


def _retrieve_name(var, offset=1):
    """Get the python variable name of the calling variable i.e. _retrieve_name(a) == 'a'"""
    caller_locals = inspect.currentframe()
    for _ in range(offset):
        caller_locals = caller_locals.f_back
    caller_locals = caller_locals.f_locals.items()
    return suffix_join([name for name, value in caller_locals if value is var], sep=', ', suffix_sep=' or ')


def pretty_print_values(*args, **kwargs):
    """Print variables and values using introspection.

    Example:
        # Instead of print(f"variable1={variable1}, variable2={variable2}, ..., variableN={variableN}")
        pretty_print_values(variable1, variable2, variableN)
    """
    names = list(map(lambda x: _retrieve_name(x, offset=3), args))
    names = {name: value for name, value in zip(names, args)}
    names.update(kwargs)
    pp_string = suffix_join([f"{name}={value}" for name, value in names.items()], sep=', ', suffix_sep=' and ')
    print(pp_string)
    return pp_string



import os
import pathlib


class Path(pathlib.Path):
    def __new__(cls, *args, **kwargs):
        if cls is Path:
            cls = WindowsPath if os.name == 'nt' else PosixPath
        self = super().__new__(cls, *args, **kwargs)
        return self

    def unique(self):
        """Ensure this path doesn't already exist, will append a number to make it unique."""
        obj = self
        unique_folder = 0
        while obj.exists():
            obj = self.parent / f"{self.stem}_{str(unique_folder)}{self.suffix}"
            unique_folder += 1
        return obj


class PosixPath(Path, pathlib.PosixPath):
    pass


class WindowsPath(Path, pathlib.WindowsPath):
    pass



from . import datasets
from . import display
from . import logging
from . import perception
from . import system
from . import time
from . import ui
from . import path
from . import metrics

from .version import version as __version__



import datetime
from abc import abstractmethod, ABC
from typing import List, Union, Optional


class CocoModes:
    """ Enum class to represent the different flavours of CocoFiles"""
    DETECTION = "detection"
    KEYPOINT = "keypoint"
    STUFF = "stuff"
    PANOPTIC = "panoptic"
    CAPTIONING = "captioning"
    modes = [DETECTION, KEYPOINT, STUFF, PANOPTIC, CAPTIONING]

    @staticmethod
    def is_valid(mode: str):
        """Returns true if mode is a valid Coco file format/flavour

        Args:
            mode: One of CocoModes.modes
        """
        return mode in CocoModes.modes


class __CocoField:
    """Abstract base class for a Coco field (a section of the json document)

    See http://cocodataset.org/#format-data
    """

    @property
    @abstractmethod
    def field_name(self):
        """Abstract base property for the name of the Coco field i.e info, licence etc."""
        pass

    def coco_json(self):
        """Returns the dict of class parameters"""
        return self.__dict__

    def __str__(self):
        """Pretty print when str() is called"""
        return self.coco_json().__str__()


class __CocoIterable:
    @abstractmethod
    def get_coco_fields(self):
        """Abstract method to get the coco field objects from the child class

        child.get_coco_fields() should return an array of __CocoField objects i.e CocoImage
        """
        pass

    def coco_json(self):
        """Returns the dictionary representation of all the Coco field classes in the iterable"""
        return [x.coco_json() for x in self.get_coco_fields()]


class __CocoSet(ABC, __CocoIterable, set):
    """Utility iterable for set"""
    pass


class __CocoDict(ABC, __CocoIterable, dict):
    """Utility iterable for dict"""
    pass


class __CocoList(ABC, __CocoIterable, list):
    """Utility iterable for list"""
    pass


class CocoInfo(__CocoField):
    """Represents the info section of the Coco format

    info{
        "year": int,
        "version": str,
        "description": str,
        "contributor": str,
        "url": str,
        "date_created": str,
    }
    """
    field_name = "info"

    def __init__(self, year: int = 0, version: str = "0.0.1", description: str = "", contributor: str = "",
            url: str = "", date_created: str = ""):
        if date_created == "":
            date_created = datetime.datetime.now().strftime("%d-%m-%Y")
        if year == 0:
            date_created = int(datetime.datetime.now().strftime("%Y"))

        self.year = year
        self.version = version
        self.description = description
        self.contributor = contributor
        self.url = url
        self.date_created = date_created


class CocoImage(__CocoField):
    """Represents the image section of the Coco format

    image{
        "id": int,
        "width": int,
        "height": int,
        "file_name": str,
        "license": int,
        "flickr_url": str,
        "coco_url": str,
        "date_captured": str,
    }
    """
    field_name = "image"

    def __init__(self, image_id: int, width: int, height: int, file_name: str, image_license: int, flickr_url: str,
            coco_url: str, date_captured: str, **kwargs):
        """
        Args:
            **kwargs: Extra information to store in the COCO document i.e depth info
        """
        self.id = image_id
        self.width = width
        self.height = height
        self.file_name = file_name
        self.license = image_license
        self.flickr_url = flickr_url
        self.coco_url = coco_url
        self.date_captured = date_captured
        for k, v in kwargs.items():
            setattr(self, k, v)


class CocoLicence(__CocoField):
    """Represents the licence section of the Coco format

    license{
        "id": int,
        "name": str,
        "url": str,
    }
    """
    field_name = "licence"

    def __init__(self, licence_id: int = 0, name: str = "", url: str = ""):
        self.id = licence_id
        self.name = name
        self.url = url


class CocoDetectionAnnotation(__CocoField):
    """Represents the annotation section of the Coco Object Detection format

    annotation{
        "id": int,
        "image_id": int,
        "category_id": int,
        "segmentation": [polygon] or RLE,
        "area": float,
        "bbox": [x,y,width,height],
        "iscrowd": 0 or 1,
    }
    """
    field_name = "annotation"

    def __init__(self, annotation_id: int, image_id: int, category_id: int, segmentation: Union[dict, list],
            area: float, bbox: List[float], iscrowd: int, **kwargs):
        self.id = annotation_id
        self.image_id = image_id
        self.category_id = category_id
        self.segmentation = segmentation
        self.area = area
        self.bbox = bbox
        self.iscrowd = iscrowd
        for k, v in kwargs.items():
            setattr(self, k, v)


class CocoCategory(__CocoField):
    """Represents the category section of the Coco Object Detection format

    category{
        "id": int,
        "name": str,
        "supercategory": str,
    }
    """
    field_name = "category"

    def __init__(self, category_id: int, name: str, supercategory: str):
        self.id = category_id
        self.name = name
        self.supercategory = supercategory


class CocoCategories(__CocoSet):
    """Utility class for generating a CocoList of categories for the Coco categories Object Detection format

    When exported it will automatically calculate category ids and super categories. Inherits from the set object.

    Attributes:
        classes: List of the classes/labels
    """
    field_name = "categories"

    def __init__(self, classes: Optional[List[str]] = None):
        if classes is None:
            classes = []
        super().__init__(classes)

    @property
    def classes_list(self):
        """Return a sorted list of class names (alphabetically a->z)"""
        return sorted(list(self))

    @property
    def classes_to_idx(self):
        """Returns a look up dict to go from class name to class id"""
        return {k: i + 1 for i, k in enumerate(self.classes_list)}

    @property
    def idx_to_classes(self):
        """Returns a look up dict to go from class id to class name"""
        return {i + 1: k for i, k in enumerate(self.classes_list)}

    @property
    def class_to_super_categories(self):
        """Returns the super categories look up dict to go from class name to super category

        Tokenises the class names and tries to find the best matching tokens from other classes, then assigns these as
        the super class
        """
        classes = self.classes_list
        classes_word_splits = {k: k.split(' ') for k in classes}
        super_categories = {k: "" for k in classes_word_splits.keys()}
        for class_name, class_tokens in classes_word_splits.items():
            super_categories_scores = {t: 0 for t in class_tokens}
            for token in class_tokens:
                token_scores = {t: 0 for t in class_tokens}

                for comparison_class in classes:
                    if comparison_class == class_name:
                        continue
                    if token in comparison_class:
                        token_scores[token] += 1

                max_score = max(token_scores.keys(), key=(lambda key: token_scores[key]))
                super_categories_scores[max_score] += token_scores[max_score]

            super_category = max(super_categories_scores.keys(), key=(lambda key: token_scores[key]))
            super_categories[class_name] = super_category

        return super_categories

    @property
    def categories(self):
        """Returns the classes as an array of CocoCategory objects"""
        idx_to_classes = self.idx_to_classes
        super_categories = self.class_to_super_categories
        return [CocoCategory(category_id=class_id, name=class_name, supercategory=super_categories[class_name])
                for class_id, class_name in idx_to_classes.items()]

    def get_coco_fields(self):
        """Returns the list of CocoCategory objects"""
        return self.categories


class CocoList(__CocoList):
    @property
    def field_name(self):
        """Tries to assume what category/field name based on existing entries"""
        if len(self) == 0:
            raise ValueError("Can't get name attribute of CocoList since it's empty")
        return self[0].field_name + "s"

    def get_coco_fields(self):
        """Returns the CocoFields"""
        return self


def combine_coco_objects(coco_objects: List[Union[__CocoField, __CocoIterable]]):
    """Combines the coco_objects array into a single dict for saving using json to coco.json files"""
    return {x.field_name: x.coco_json() for x in coco_objects}



import pathlib

import setuptools

__package_name = "raytils"
__root_directory_path = pathlib.Path(__file__).parent
__package_directory_path = __root_directory_path / __package_name
__version_file_path = __package_directory_path / "version.py"


def get_version():
    import re
    with open(str(__version_file_path), "rb") as fh:
        regex = r"\s*([\d.]+)"
        matches = list(re.finditer(regex, str(fh.readline()), re.MULTILINE))
        if len(matches) != 1:
            raise ValueError("Version not in '{}' file, this file must only be one line of version = 'X.X.X'".format(
                str(__version_file_path)))
        version = matches[0].group()
        return version


if __name__ == '__main__':
    __version__ = get_version()

    setuptools.setup(
        name='raytils',
        version=__version__,
        author="Raymond Tunstill",
        author_email="ray.tunstill@live.co.uk",
        description="Suite of utilities created by Raymond Tunstill.",
        long_description=open('README.md').read(),
        long_description_content_type="text/markdown",
        url="https://raymondkirk.github.io/raytils/",
        licence="GNU GPLv3",
        packages=list(filter(lambda x: __package_name in x, setuptools.find_packages(exclude=["tests"]))),
        classifiers=[
            "Programming Language :: Python :: 3",
        ],
        setup_requires=['wheel', "Cython"],
        install_requires=[
            'Cython',
            'numpy',
            'scikit-image',
            'scipy',
            'opencv-python',
            'Pillow',
            'tqdm',
            'rays_pycocotools',
            'requests',
            'imagesize',
            'pandas',
            'scikit-learn',
            'plotly',
        ],
        python_requires='>=3.6',
    )



#!/usr/bin/env python

#  Raymond Kirk (Tunstill) Copyright (c) 2020
#  Email: ray.tunstill@gmail.com

# LoadBalancing utilities from raytils (https://github.com/RaymondKirk/raytils, https://pypi.org/project/raytils/)
from typing import Callable, Union, List, Tuple

from raytils.system import available_cpu_count
from raytils.time import FPSCounter

import queue

import threading

from timeit import default_timer as timer


class Worker(threading.Thread):
    """Runs jobs from LoadBalancer.

    Work from the LoadBalancer should be in the form of [func, args, kwargs].
    func should take a parameter job_meta containing specifics about the worker running the job.
    """
    def __init__(self, q: queue.Queue, worker_id: int, *args, **kwargs):
        self.q = q
        self.worker_id = worker_id
        self.data_retrieval_rate = FPSCounter(q.maxsize)
        self.processing_rate = FPSCounter(q.maxsize)
        super(Worker, self).__init__(*args, **kwargs)

    def run(self):
        while True:
            try:
                self.data_retrieval_rate.tic()
                work = self.q.get()
                self.data_retrieval_rate.toc()
            except queue.Empty:
                return

            # Create dictionary of meta values to pass to the worker
            job_meta = {
                "worker_id": self.worker_id,  # This is my id
                "worker_data_retrieval_rate": self.data_retrieval_rate.get_fps(),  # I get data every n times per second
                "worker_job_processing_rate": self.processing_rate.get_fps(),  # I process data every n times per second
            }

            self.processing_rate.tic()
            work[0](*work[1], job_meta=job_meta, **work[2])
            self.processing_rate.toc()
            self.q.task_done()


class LoadBalancer(queue.Queue):
    """Add jobs in the form of (func, args, kwargs) to a task queue and automatically scales number of workers.

    Usage:

    .. code-block:: python

        from raytils.system import LoadBalancer
        # We will add heavy tasks to the worker to not block the main thread
        workers = LoadBalancer(maxsize=30, threads=8)

        # Your function must have a job_meta keyword argument so your workers
        # can pass important information about themselves
        def save_dict(file_path, data, job_meta):
            import json
            with open(file_path, "w") as fh:
                json.dump(data, fh)

        # Save 10000 dictionaries to files to demonstrate a heavy load
        for idx in range(10000):
            dict_to_save = {i: i for i in range(10000)}
            file_path = f"{idx}.json"
            # Serially we would do save_dict(file_path, dict_to_save)
            # However we can do it much faster
            workers.add_task(save_dict, [file_path, dict_to_save])
    """
    def __init__(self, maxsize: int = 0, threads: int = 1, auto: bool = True):
        """
        Example:
        Args:
            maxsize: Maximum size of the created jobs queue. If maxsize is <= 0, the queue size is infinite.
            threads: Number of workers to start threads for
            auto: If true workers will be automatically created if the demand is too high.
        """
        super(LoadBalancer, self).__init__(maxsize=maxsize)
        self.max_threads = available_cpu_count()
        self.auto = auto
        self.number_of_threads = 0
        self.tasks_per_second = FPSCounter(queue_length=maxsize)
        # Number of seconds to allow queue to catch up before adding more cores
        self._thread_grace_period = 2.0  # TODO: these hard-coded values should be tunable
        self._last_thread_added = timer()
        self._add_thread_threshold_percent = 0.8  # TODO: these hard-coded values should be tunable
        for _ in range(threads):
            self._add_thread(check_grace_period=False)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.join()

    def _add_thread(self, check_grace_period: bool = True):
        """Function that's used internally. If the queue of jobs ever becomes full a new thread is spawned to try to
        handle the new information in realtime"""
        # Do not add more threads if the CPU cannot support it
        if self.number_of_threads >= self.max_threads:
            return

        # Grace period so cores don't quickly accelerate to system maximum
        if check_grace_period and (timer() - self._last_thread_added) <= self._thread_grace_period:
            return

        Worker(self, self.number_of_threads).start()
        self.number_of_threads += 1
        self._last_thread_added = timer()

    def add_task(self, func: Callable, args: Union[List, Tuple] = None, kwargs: dict = None, wait: bool = False):
        """Add a task to be processed by one of the available workers

        Args:
            func: callable of the signature func(*args, **kwargs) or func(arg1, arg2, ..., kwarg1, job_meta=None)
            args: list of arguments to pass to the func calling func(*args, job_meta={...}, **kwargs)
            kwargs: list of keyword arguments to pass to the func calling func(*args, job_meta={...}, **kwargs)
            wait: If true blocks until the item is placed in the queue

        Returns:
            bool: True if item placed in the queue otherwise false
        """
        # If the number of tasks_per_second being added is lower than the current approx queue size
        # then add a thread to deal with the extra work
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        self.tasks_per_second.toc()
        self.tasks_per_second.tic()

        # Add a thread if queue_size growing faster than current processing or is nearing capacity
        if self.auto:
            queue_size = self.qsize()
            queue_capacity = queue_size / self.maxsize
            if queue_size >= self.tasks_per_second.get_fps() or queue_capacity >= self._add_thread_threshold_percent:
                self._add_thread()

        try:
            if wait:
                self.put((func, args, kwargs))
            else:
                self.put_nowait((func, args, kwargs))
        except queue.Full:
            if self.auto:
                # If the queue becomes full add a thread to balance the workload
                self._add_thread()  # attempt to add a thread to balance the workload
            return False

        return True



import copy
import math
import os
import time
from collections import OrderedDict
from typing import Tuple, Optional, List

import matplotlib.pyplot as plt
import numpy as np
from rays_pycocotools.coco import COCO
from rays_pycocotools.cocoeval import COCOeval as __COCOeval
from rays_pycocotools.cocoeval import Params as __Params

__all__ = ["COCOEvalExtended"]


class ParamsExtended(__Params):
    """ Adds the ability to define custom IoU (Intersection of Union) ranges for COCO evaluation (pycocotools.cocoeval)

    Extends the pycocotools.cocoeval.Params object.
    Best used with https://github.com/RaymondKirk/cocoapi/tree/python-api instead of the cocodataset/cocoapi package

    Distribution for setup.py or pip 'pycocotools @ https://github.com/RaymondKirk/cocoapi/archive/python-api.zip'

    Attributes:
        iou_type: 'segm' or 'bbox' determines how the IoU is calculated (mask or bounding box)
        iou_rng: The range of IoU values for evaluation in the form (IoU min, IoU max, IoU step) the default parameters
                 are (0.5, 0.95, 0.05)
        rec_rng: The range of Recall values to calculate precision at for evaluation in the form
                 (recall min, recall max, recall step) the default parameters are (0.0, 1.0, 0.01)
        area_rng: The range of area values to threshold the evaluation at. Default (0->Inf, 0-32^2, 32^2-96^2, 96^2-Inf)
        area_rng_labels: Descriptor of area range. Default ('all', 'small', 'medium', 'large').
    """

    def __init__(self, iou_type: str = 'segm', iou_rng: Tuple[float, float, float] = None,
            rec_rng: Tuple[float, float, float] = None,
            area_rng: Tuple[List[float], List[float], List[float], List[float]] = None,
            area_rng_labels: Tuple[str, str, str, str] = None):
        super().__init__(iou_type)

        # Override iou parameters set by set[Det/Kp]Params function
        if iou_rng is None:
            iou_rng = (0.5, 0.95, 0.05)
        if rec_rng is None:
            rec_rng = (0.0, 1.0, 0.01)
        if area_rng_labels is None:
            area_rng_labels = ('all', 'small', 'medium', 'large')
        if area_rng is None:
            area_rng = ([0 ** 2, float("inf")], [0 ** 2, 32 ** 2], [32 ** 2, 96 ** 2], [96 ** 2, float("inf")])
        if 0 <= len(area_rng) > 4:
            raise ValueError("Invalid area range defined")
        self.iouLow, self.iouHigh, self.iouStep = iou_rng
        self.recLow, self.recHigh, self.recStep = rec_rng
        self.iouThrs = np.linspace(self.iouLow, self.iouHigh,
                                   int(np.round((self.iouHigh - self.iouLow) / self.iouStep)) + 1, endpoint=True)
        self.recThrs = np.linspace(self.recLow, self.recHigh,
                                   int(np.round((self.recHigh - self.recLow) / self.recStep)) + 1, endpoint=True)
        self.areaRng = list(area_rng)
        self.areaRngLbl = list(area_rng_labels)
        print()


class COCOEvalExtended(__COCOeval):
    """Extends the COCOeval object in pycocotools.cocoeval

    COCOEval computes the precision and recall over the dependant variable Iou Thresh
    Provides additional functionality such as the abilty to generate error analysis graphs.

    The analysis function is in the style of http://dhoiem.cs.illinois.edu/publications/eccv2012_detanalysis_derek.pdf

    Attributes:    #  cocoGt=..., cocoDt=...       # load dataset and results
        coco_gt: COCO json dataset ground truth file (gt=ground truth)
        coco_dt: COCO json dataset predictions file (dt=detections)
        iou_type: 'segm' or 'bbox' determines how the IoU is calculated (mask or bounding box)
        iou_rng: The range of IoU values for evacuation in the form (IoU min, IoU max, IoU step) the default parameters
                 are (0.5, 0.95, 0.05)
        rec_rng: The range of Recall values to calculate precision at for evaluation in the form
                 (recall min, recall max, recall step) the default parameters are (0.0, 1.0, 0.01)
        area_rng: The range of area values to threshold the evaluation at. Default (0->Inf, 0-32^2, 32^2-96^2, 96^2-Inf)
        area_rng_labels: Descriptor of area range. Default ('all', 'small', 'medium', 'large').
    """

    def __init__(self, coco_gt: COCO = None, coco_dt: COCO = None, iou_type: str = 'segm',
            iou_rng: Optional[Tuple[float, float, float]] = None,
            rec_rng: Tuple[float, float, float] = None,
            area_rng: Tuple[List[float], List[float], List[float], List[float]] = None,
            area_rng_labels: Tuple[str, str, str, str] = None):
        super().__init__(coco_gt, coco_dt, iou_type)
        self.analyze_figures = {}  # for analyse function to store plots
        self.params = ParamsExtended(iou_type, iou_rng, rec_rng, area_rng, area_rng_labels)
        if coco_gt is not None:
            self.params.imgIds = sorted(coco_gt.getImgIds())
            self.params.catIds = sorted(coco_gt.getCatIds())
        self.eval_cm = {}
        self.eval_images_cm = {}
        self.ious_cm = {}

    @staticmethod
    def from_file(coco_gt_json, coco_dt_json, *args, **kwargs):
        """Utility function to get COCOEvaluator object from two file paths instead of COCO objects

        Args:
            coco_gt_json: File path to the *.json coco ground truth file
            coco_dt_json: File path to the *.json coco predictions file
            *args: Passed to the constructed COCOEvalExtended object
            **kwargs: Passed to the constructed COCOEvalExtended object

        Returns:
            COCOEvalExtended object
        """
        coco_gt = COCO(coco_gt_json)
        coco_dt = coco_gt.loadRes(coco_dt_json)
        return COCOEvalExtended(coco_gt, coco_dt, *args, **kwargs)

    def summarise_extended(self):
        """Replaces functionality of COCOEval.summarize with added evaluation metrics

            Adds clarity to an otherwise unreadable function.
        """

        def __summarise(metric, iou_threshold=None, area_range_str='all', max_detections=100):
            p = self.params
            i_str = ' {:<18} {} @[ IoU={:<9} | area={:>6s} | maxDets={:>3d} ] = {:0.3f}'
            title_str = {'af1': 'Average F1', 'ap': 'Average Precision', 'ar': 'Averge Recall'}[metric]
            type_str = {'af1': '(AF1)', 'ap': '(AP) ', 'ar': '(AR) '}[metric]

            if iou_threshold is None:
                iou_str = '{:0.2f}:{:0.2f}'.format(p.iouThrs[0], p.iouThrs[-1])
                iou_threshold_params_index = slice(None)  # All the iou thresholds
            else:
                iou_str = '{:0.2f}'.format(iou_threshold)
                iou_threshold_params_index = np.where(iou_threshold == p.iouThrs)[0]

            area_index = [i for i, area_range_lbl in enumerate(p.areaRngLbl) if area_range_lbl == area_range_str]
            max_dets_index = [i for i, max_dets in enumerate(p.maxDets) if max_dets == max_detections]

            # [TxRxKxAxM] T=p.iou_threshes, R=p.recall_threshes, K=classes, A=p.areas, M=p.max dets
            if metric == 'ap':
                # Dimension of precision: [T x R x K x A x M] (see above notation)
                s_value = self.eval['precision'][iou_threshold_params_index][:, :, :, area_index, max_dets_index]
            elif metric == 'ar':
                # Dimension of recall: [T x K x A x M] (see above notation)
                s_value = self.eval['recall'][iou_threshold_params_index][:, :, area_index, max_dets_index]
            elif metric == "af1":
                precision = self.eval['precision'][iou_threshold_params_index]
                recall = self.eval['recall'][iou_threshold_params_index]
                precision = precision.mean(axis=1)  # Mean on the recall threshold axis R
                f1 = 2 * (precision * recall) / (precision + recall + np.finfo(float).eps)
                s_value = f1[:, :, area_index, max_dets_index]
            else:
                raise ValueError(f"Metric '{metric}' isn't supported in COCOEvalExtended")

            # Check all values are > -1 (-1 used for invalid/NaN)
            if len(s_value[s_value > -1]) == 0:
                metric_value = -1
            else:
                metric_value = np.mean(s_value[s_value > -1])

            print(i_str.format(title_str, type_str, iou_str, area_range_str, max_detections, metric_value))
            return metric_value

        def __summarise_detections():
            stats = np.zeros((18,))
            stats[0] = __summarise("ap")
            stats[1] = __summarise("ap", iou_threshold=.5, max_detections=self.params.maxDets[2])
            stats[2] = __summarise("ap", iou_threshold=.75, max_detections=self.params.maxDets[2])
            stats[3] = __summarise("ap", area_range_str='small', max_detections=self.params.maxDets[2])
            stats[4] = __summarise("ap", area_range_str='medium', max_detections=self.params.maxDets[2])
            stats[5] = __summarise("ap", area_range_str='large', max_detections=self.params.maxDets[2])
            stats[6] = __summarise("ar", max_detections=self.params.maxDets[0])
            stats[7] = __summarise("ar", max_detections=self.params.maxDets[1])
            stats[8] = __summarise("ar", max_detections=self.params.maxDets[2])
            stats[9] = __summarise("ar", area_range_str='small', max_detections=self.params.maxDets[2])
            stats[10] = __summarise("ar", area_range_str='medium', max_detections=self.params.maxDets[2])
            stats[11] = __summarise("ar", area_range_str='large', max_detections=self.params.maxDets[2])

            stats[12] = __summarise("af1")
            stats[13] = __summarise("af1", iou_threshold=.5, max_detections=self.params.maxDets[2])
            stats[14] = __summarise("af1", iou_threshold=.75, max_detections=self.params.maxDets[2])
            stats[15] = __summarise("af1", area_range_str='small', max_detections=self.params.maxDets[2])
            stats[16] = __summarise("af1", area_range_str='medium', max_detections=self.params.maxDets[2])
            stats[17] = __summarise("af1", area_range_str='large', max_detections=self.params.maxDets[2])

            return stats

        def __summarise_keypoints():
            stats = np.zeros((10,))
            stats[0] = __summarise(1, max_detections=20)
            stats[1] = __summarise(1, max_detections=20, iou_threshold=.5)
            stats[2] = __summarise(1, max_detections=20, iou_threshold=.75)
            stats[3] = __summarise(1, max_detections=20, area_range_str='medium')
            stats[4] = __summarise(1, max_detections=20, area_range_str='large')
            stats[5] = __summarise(0, max_detections=20)
            stats[6] = __summarise(0, max_detections=20, iou_threshold=.5)
            stats[7] = __summarise(0, max_detections=20, iou_threshold=.75)
            stats[8] = __summarise(0, max_detections=20, area_range_str='medium')
            stats[9] = __summarise(0, max_detections=20, area_range_str='large')
            return stats

        def __summarise_detections_confusion_matrix():
            iou_ind = area_ind = 0
            max_det_ind = len(self.params.maxDets) - 1
            cm = self.eval_cm["confusion_matrix"][iou_ind, area_ind, max_det_ind]
            class_names = ["None"] + [c["name"] for c in self.cocoGt.cats.values()]
            cm_printer = ConfusionMatrix(cm, class_names=class_names)

            i_str = ' {:<18} @[ IoU={:<9} | area={:>6s} | maxDets={:>3d} ]'
            iou_str = '{:0.2f}'.format(self.params.iouThrs[iou_ind])
            area_str = self.params.areaRngLbl[area_ind]
            print(i_str.format("\nConfusion matrix", iou_str, area_str, self.params.maxDets[max_det_ind]))
            cm_printer.tabulate()
            cm_printer.summarise()
            # print(i_str.format("\nBinary Confusion matrix", iou_str, area_str, self.params.maxDets[max_det_ind]))
            # bcm = self.eval_cm["binary_confusion_matrix"][iou_ind, area_ind, max_det_ind]
            # bcm_class_names = ["None"] + ['+'.join([c["name"] for c in self.cocoGt.cats.values()])]
            # bcm_printer = ConfusionMatrix(bcm, class_names=bcm_class_names)
            # bcm_printer.tabulate()
            # bcm_printer.summarise()

        if not self.eval:
            raise Exception('Please run accumulate() first')

        iou_type = self.params.iouType
        if iou_type == 'segm' or iou_type == 'bbox':
            if self.eval_cm:
                __summarise_detections_confusion_matrix()
            summarise = __summarise_detections
        elif iou_type == 'keypoints':
            summarise = __summarise_keypoints
        else:
            raise ValueError(f"IoU type '{iou_type}' not supported in COCOEvalExtended")

        self.stats = summarise()

    def __str__(self):
        self.summarise_extended()

    def accumulate(self, p=None, cm=True):
        """Wrapper for COCOEval.accumulate to also run extra metrics"""
        super().accumulate(p)
        if cm:
            self.__accumulate_class_wise_confusion_matrix()

    def evaluate(self, cm=True):
        """Wrapper for COCOEval.evaluate to ensure extra metrics are computed (confusion matrix)

            Summary of internal cocoapi metrics computation here:
            https://blog.zenggyu.com/en/post/2018-12-16/an-introduction-to-evaluation-metrics-for-object-detection/
        Args:
            cm: Confusion matrix if true in self.eval_cm if true
        """
        super().evaluate()

        if cm:
            p = self.params
            if p.iouType == 'segm' or p.iouType == 'bbox':
                print('Running per image evaluation for confusion matrix...')
                tic = time.time()

                # loop through images, area range, max detection number
                category_ids = p.catIds if p.useCats else [-1]
                compute_iou = self.computeIoU
                max_det = p.maxDets[-1]
                temp_use_cats = self.params.useCats
                self.params.useCats = False
                self.ious_cm = {(imgId, catId): compute_iou(imgId, catId) for imgId in p.imgIds for catId in
                                category_ids}
                evaluate_image_confusion_matrix = self.__evaluate_image_confusion_matrix
                self.eval_images_cm = [evaluate_image_confusion_matrix(imgId, areaRng, max_det)
                                       for areaRng in p.areaRng
                                       for imgId in p.imgIds]
                self.params.useCats = temp_use_cats
                toc = time.time()
                print('DONE (t={:0.2f}s).'.format(toc - tic))
            else:
                print("Warning: Only 'segm' and 'bbox' is currently supported for 'p.iouType' confusion matrix")

    def __evaluate_image_confusion_matrix(self, img_id, a_rng, max_det):
        """Perform confusion matrix evaluation for single image. Gather all actual and predicted values.

            1. Generates a y_true array with the classes of ground truth annotations, skipping annotations
                outside the range of the threshold parameters
            2. Matches detections to the ground truth annotations and generates a y_pred array of detection classes
            3. For all unmatched detections update y_true with 0 and y_pred with the class ID
            4. Return y_true and y_pred

        """
        params = self.params
        ground_truth = [_ for cId in params.catIds for _ in self._gts[img_id, cId]]
        detections = [_ for cId in params.catIds for _ in self._dts[img_id, cId]]
        if len(ground_truth) == 0 and len(detections) == 0:
            return None

        for gt_data in ground_truth:
            gt_data['_ignore'] = gt_data['ignore'] or (gt_data['area'] < a_rng[0] or gt_data['area'] > a_rng[1])

        # Sort detections with the highest score first and ground truth with ignore flag last
        gt_sorted_indexes = np.argsort([g['_ignore'] for g in ground_truth], kind='mergesort')
        ground_truth = [ground_truth[i] for i in gt_sorted_indexes]
        dt_sorted_indexes = np.argsort([-d['score'] for d in detections], kind='mergesort')
        detections = [detections[i] for i in dt_sorted_indexes[0:max_det]]
        gt_iscrowd = [int(o['iscrowd']) for o in ground_truth]

        # load computed ious
        ious = self.ious_cm[img_id, params.catIds[0]][:, gt_sorted_indexes] \
            if len(self.ious_cm[img_id, params.catIds[0]]) > 0 else self.ious_cm[img_id, params.catIds[0]]

        iou_thresh_count = len(params.iouThrs)

        ground_truth_matches = np.zeros((iou_thresh_count, len(ground_truth)))
        ground_truth_ignore = np.array([g['_ignore'] for g in ground_truth])
        eps = np.finfo(float).eps

        y_true = np.asarray(
            [gt["category_id"] for i, gt in enumerate(ground_truth) if not ground_truth_ignore[i]]
        ).reshape(1, -1).repeat(iou_thresh_count, 0).tolist()
        y_pred = np.zeros_like(y_true).tolist()

        if not len(ious) == 0:
            for iou_thresh_ind, iou_thresh in enumerate(params.iouThrs):
                unmatched_y_pred = []
                for detection_ind, d in enumerate(detections):
                    # Skip if invalid area range
                    if d['area'] < a_rng[0] or d['area'] > a_rng[1]:
                        continue

                    # Information about best match so far (m=-1 -> unmatched)
                    iou = min([iou_thresh, 1 - eps])
                    gt_match_index = -1

                    for gt_index in range(len(ground_truth)):
                        # If this ground truth detection already matched and not a crowd then continue
                        if ground_truth_matches[iou_thresh_ind, gt_index] > 0 and not gt_iscrowd[gt_index]:
                            continue
                        # If dt matched to reg gt, and on ignore gt, stop
                        if gt_match_index > -1 and ground_truth_ignore[gt_match_index] == 0 and ground_truth_ignore[
                            gt_index] == 1:
                            break
                        # Continue to next gt unless better match made
                        if ious[detection_ind, gt_index] < iou:
                            continue
                        # If match successful and best so far, store appropriately
                        iou = ious[detection_ind, gt_index]
                        gt_match_index = gt_index

                    # If no match then continue else store id of match for both dt and gt
                    if gt_match_index == -1:
                        unmatched_y_pred.append(d["category_id"])
                        continue

                    if not ground_truth_ignore[gt_match_index]:
                        y_pred[iou_thresh_ind][gt_match_index] = d["category_id"]
                    ground_truth_matches[iou_thresh_ind, gt_match_index] = d['id']

                y_true[iou_thresh_ind].extend([0] * len(unmatched_y_pred))
                y_pred[iou_thresh_ind].extend(unmatched_y_pred)

        # Store results for given image
        return {
            'image_id': img_id,
            'aRng': a_rng,
            'maxDet': max_det,
            'y_true': y_true,
            'y_pred': y_pred,
        }

    def __accumulate_class_wise_confusion_matrix(self, p=None):
        '''Accumulate per image evaluation results for confusion matrix and store the result in self.eval

        :param p: input params for evaluation
        :return: None
        '''
        print('Accumulating evaluation results for confusion matrix...')
        tic = time.time()

        if not self.eval_images_cm:
            print('Please run evaluate() first')

        # allows input customized parameters
        if p is None:
            p = self.params

        p.catIds = p.catIds if p.useCats == 1 else [-1]
        iou_len = len(p.iouThrs)
        class_len = len(p.catIds) if p.useCats else 1
        area_len = len(p.areaRng)
        mdets_len = len(p.maxDets)

        # [TxKxAxMxCxC] (C=N Classes)
        n_classes = class_len + 1
        class_wise_confusion_matrix = -np.ones((iou_len, area_len, mdets_len, n_classes, n_classes))
        binary_confusion_matrix = -np.ones((iou_len, area_len, mdets_len, 2, 2))

        # Create dictionaries for future indexing
        _pe = self._paramsEval
        catIds = _pe.catIds if _pe.useCats else [-1]
        setK = set(catIds)
        setA = set(map(tuple, _pe.areaRng))
        setM = set(_pe.maxDets)
        setI = set(_pe.imgIds)

        # Get index lists of parameters to evaluate
        if 0 in setK:
            raise ValueError("Please start category ids from 1! Cannot assign 0 for No Object class")
        class_ind_list = [0] + [int(k) for n, k in enumerate(p.catIds) if k in setK]
        m_list = [m for n, m in enumerate(p.maxDets) if m in setM]
        area_ind_list = [n for n, a in enumerate(map(lambda x: tuple(x), p.areaRng)) if a in setA]
        image_ind_list = [n for n, i in enumerate(p.imgIds) if i in setI]
        image_list_count = len(_pe.imgIds)

        # Retrieve evaluation image at each area range and max number of detections
        for area_ind in area_ind_list:
            Na = area_ind * image_list_count
            for max_det_ind, max_det in enumerate(m_list):
                # Get all evaluation images for the current parameter combination
                evaluation_images = [self.eval_images_cm[Na + i] for i in image_ind_list]
                evaluation_images = [e for e in evaluation_images if e is not None]
                if len(evaluation_images) == 0:
                    continue

                y_true = [[] for _ in range(iou_len)]
                y_pred = [[] for _ in range(iou_len)]

                for img in evaluation_images:
                    for iou_thresh in range(iou_len):
                        y_true[iou_thresh].extend(img["y_true"][iou_thresh][0:max_det])
                        y_pred[iou_thresh].extend(img["y_pred"][iou_thresh][0:max_det])

                class_wise_confusion_matrix[:, area_ind, max_det_ind, :, :] = [
                    ConfusionMatrix.get_matrix(gt, dt, labels=class_ind_list) for gt, dt in zip(y_true, y_pred)
                ]

                binary_confusion_matrix[:, area_ind, max_det_ind, :, :] = [
                    ConfusionMatrix.get_matrix(
                        [0 if g == 0 else 1 for g in gt],
                        [0 if d == 0 else 1 for d in dt], labels=[0, 1]) for gt, dt in zip(y_true, y_pred)
                ]

        self.eval_cm = {'params': p, 'confusion_matrix': class_wise_confusion_matrix,
                        "binary_confusion_matrix": binary_confusion_matrix}

        toc = time.time()
        print('DONE (t={:0.2f}s).'.format(toc - tic))

    def plot_class_pr(self, fig=None, save_to_dir=None, show=False, save_ext=".pdf"):
        """Display the precision/recall curve of the COCOeval evaluation

        This function shows a basic PR curve for the COCOeval evaluation

        Usage:
        .. code-block:: python

            coco_evaluator = COCOEvalExtended.from_file(ground_truth_file_path, predictions_file_path)
            coco_evaluator.plot_class_pr(class_names=coco_evaluator.cocoGt.cats, save_to_dir=None, show=True)

        Args:
            fig: A figure object to plot the pr curve on, if None it will create one
            class_names: The list of class names from the evaluated predictions
            save_to_dir: The directory to save the plots to i.e 'figures', if None the figures are not saved
            show: Opens the plots in a matplotlib window if True
            save_ext: matplotlib figure save extension (.pdf|.png|.jpg etc.)

        Returns:
            The figure object of the rendered plot
        """

        def __plot_pr_curve(iou_threshold=None, area_range_str='all', max_detections=100, fig=None, save_dir=None,
                show=False):

            # If iou_threshold == None get all the iou_thresholds
            iou_threshold_index = slice(None) if iou_threshold is None else \
                np.where(iou_threshold == self.params.iouThrs)[0]

            # Get the area index in the params that corresponds to the area string
            area_index = self.params.areaRngLbl.index(area_range_str)
            max_det_index = self.params.maxDets.index(max_detections)

            # Get precision values for recall thresholds of self.params.recThrs
            # T=p.iou_threshes, R=p.recall_threshes, K=classes, A=p.areas, M=p.max dets
            recall_thresholds = self.params.recThrs

            precision = self.eval['precision']  # [T x R x K x A x M]
            recall = self.eval['recall']  # [T x K x A x M]

            # Calculate PR for classes
            class_id_to_index = {v: i for i, v in enumerate(self.eval["params"].catIds)}
            classes = {lbl_index: self.cocoGt.cats[lbl_id]["name"] for lbl_id, lbl_index in class_id_to_index.items()}

            fig = plt.figure(figsize=(9, 7))
            colours = plt.rcParams['axes.prop_cycle'].by_key()['color']

            for class_index, class_name in classes.items():
                # Get precision values for filters and class K=class_index (remove invalid values -1)
                precision_values = precision[iou_threshold_index, :, class_index, area_index, max_det_index]
                precision_values = precision_values[precision_values > -1]

                # Get recall value for filters and class K=class_index (remove invalid values -1)
                recall_value = recall[iou_threshold_index, class_index, area_index, max_det_index]
                recall_value = recall_value[recall_value > -1]

                # Get F1 Score from mean of precision and recall
                recall_mean, precision_mean = recall_value.mean(), precision_values.mean()
                f1_value = 2 * (precision_mean * recall_mean) / (precision_mean + recall_mean + np.finfo(float).eps)

                # Plot F1 chart of precision calculated at recall values
                cls_colour = colours[class_index]
                plt.step(recall_thresholds, precision_values, color=cls_colour, alpha=1.0, where='pre',
                         label=class_name)
                plt.fill_between(recall_thresholds, precision_values, alpha=0.2, color=cls_colour, step='pre')
                plt.xlabel('Recall')
                plt.ylabel('Precision')
                plt.ylim([0.0, 1.0])
                plt.xlim([0.0, 1.0])
                # Fancy F1 display (solution lines)
                for f_score in np.linspace(0.2, 0.8, num=4):
                    x = np.linspace(0.01, 1)
                    y = f_score * x / (2 * x - f_score)
                    plt.plot(x[y >= 0], y[y >= 0], color='gray', alpha=0.2)
                    plt.annotate(r'$F_1$' + '= {0:0.1f}'.format(f_score), xy=(0.89, y[45] + 0.02), color='gray')
                # Plot actual F1 score (get from array to pixel perfect plot, actual value has greater resolution)
                f1_plotted_line = 2 * ((recall_thresholds * precision_values) /
                                       (precision_values + recall_thresholds + np.spacing(1)))
                f1_idx = f1_plotted_line.argmax()
                plt.scatter(recall_thresholds[f1_idx], precision_values[f1_idx], marker='s', color=cls_colour)
                plt.annotate(r'$F_1$' + "= {:0.3f}".format(f1_value),
                             xy=(recall_thresholds[f1_idx], precision_values[f1_idx]))
                # Plot F1 solution space of F1 score
                f1_possible_x = np.linspace(0.01, 1, 100)
                f1_possible_y = f1_plotted_line[f1_idx] * f1_possible_x / (2 * f1_possible_x - f1_plotted_line[f1_idx])
                plt.plot(f1_possible_x[f1_possible_y >= 0], f1_possible_y[f1_possible_y >= 0], color=cls_colour,
                         alpha=0.5)

            iou_str = '{:0.2f}:{:0.2f}'.format(self.params.iouThrs[0], self.params.iouThrs[-1]) \
                if iou_threshold is None else '{:0.2f}'.format(iou_threshold)
            plot_name = "AllClasses_IOU={}_Area={}_MaxDets={}".format(iou_str, area_range_str, max_detections)
            plt.legend(loc='lower left')
            plt.title(plot_name.replace("_", " "))
            plt.tight_layout()

            if show:
                plt.show()

            # if save_dir is set, figure will be saved to save_dir
            if save_dir is not None:
                if not os.path.exists(save_dir):
                    os.makedirs(save_dir)

                save_file = f"{save_dir}/{plot_name.replace(' ', '_')}{save_ext}"
                print('save figure to %s' % save_file)
                fig.savefig(save_file)
            plt.close()

            return fig

        def _plotDets(fig=None, save_to_dir=None, show=False):
            return __plot_pr_curve(iou_threshold=self.params.iouLow, max_detections=self.params.maxDets[-1], fig=fig,
                                   save_dir=save_to_dir, show=show)

        if not self.eval:
            raise Exception('Please run accumulate() first')

        if self.params.iouType == 'segm' or self.params.iouType == 'bbox':
            plot_fn = _plotDets
        else:
            raise ValueError("Only iouType segm or bbox supported")

        return plot_fn(fig=fig, save_to_dir=save_to_dir, show=show)

    def __makeplot(self, recThrs, precisions, recalls, name, save_dir=None, show=False, iou_labels=None,
            font_size=None, save_ext=".pdf"):
        def strip_tfloat(f, p):  # Strips leading 0 and displays to p sig fig
            return ("{:." + str(p) + "f}").format(f)[1:]

        if iou_labels is None:
            iou_labels = ['C75', 'C50', 'LOC']
        if save_dir is not None and not os.path.exists(save_dir):
            os.makedirs(save_dir)
        # precisions  7x101x4x1]
        assert precisions.shape[2] == 4
        assert recalls.shape[1] == 4
        areaNms = ['All Areas', 'Small Area <32x32px', 'Medium Area [<96x96px]', 'Large Area [>96x96px]']
        areaNms = [
            f"{self.params.areaRngLbl[i]} area {self.params.areaRng[i][0] ** 0.5}< a < {self.params.areaRng[i][1] ** 0.5}"
            for i in range(len(self.params.areaRng))]
        figures_np = {}

        if font_size is not None:
            plt.rcParams.update({'font.size': 20})

        for aidx in range(precisions.shape[2]):
            plotname = '%s - %s' % (name, areaNms[aidx])
            # ps shape 7x101
            ps = precisions[:, :, aidx, 0].squeeze()
            rs = recalls[:, aidx, 0].squeeze()
            ap = np.mean(ps, axis=1)

            # TODO: Enable functionality to calculate f1 from recallThrs or actual recall
            # f1_value = 2 * (precision_mean * recall_mean) / (precision_mean + recall_mean + np.finfo(float).eps)
            f1_values = 2 * (ap * rs) / (ap + rs + np.finfo(float).eps)

            ds = np.concatenate((np.expand_dims(ps[0], axis=0), np.diff(ps, axis=0)))
            colors = [[.13, .54, .68], [0.5, 0.5, 0.5], [.05, .13, .29], [.75, .31, .30], [.36, .90, .38],
                      [.50, .39, .64], [1.0, .6, 0.0]]
            # Calculate each lines F1 score

            # format legend str
            ln = [iou_labels[0], iou_labels[1], iou_labels[2], 'SIM', 'CLS', 'BGR', 'FIN']
            ls = copy.copy(ln)
            for i in range(precisions.shape[0]):
                if ap[i] == 1.0:
                    ls[i] = r'AP=1.00 $F_1$=1.00 ' + ls[i]
                else:
                    ls[i] = "AP={} $F_1$={} {}".format(strip_tfloat(ap[i], 3), strip_tfloat(f1_values[i], 3), ls[i])

            fig = plt.figure(figsize=(9, 7))
            plt.stackplot(recThrs, ds, labels=ls, colors=colors, alpha=0.5)

            for f_score in np.linspace(0.2, 0.8, num=4):
                x = np.linspace(0.01, 1)
                y = f_score * x / (2 * x - f_score)
                plt.plot(x[y >= 0], y[y >= 0], color='gray', alpha=0.3)
                plt.annotate(r'$F_1$' + '= {0:0.1f}'.format(f_score), xy=(.875, y[45] + 0.02), color='gray')

            # Calculate F1 for Arrays
            plt_f1_for = [0, 1]
            for j in plt_f1_for:
                f1_plotted_line = 2 * ((recThrs * ds[j]) /
                                       (ds[j] + recThrs + np.spacing(1)))
                f1_idx = f1_plotted_line.argmax()
                plt.scatter(recThrs[f1_idx], ds[j][f1_idx], marker='s', color=colors[j])
                plt.annotate(r'$F_1$' + "= {:0.3f}".format(f1_values[j]),
                             xy=(recThrs[f1_idx], recThrs[f1_idx]))
                # Plot F1 solution space of F1 score
                f1_possible_x = np.linspace(0.01, 1, 100)
                f1_possible_y = f1_plotted_line[f1_idx] * f1_possible_x / (2 * f1_possible_x - f1_plotted_line[f1_idx])
                plt.plot(f1_possible_x[f1_possible_y >= 0], f1_possible_y[f1_possible_y >= 0], color=colors[j],
                         alpha=0.5)

            plt.axis([0.0, 1.0, 0.0, 1.0])
            plt.legend(loc='lower left')
            plt.xlabel('Recall')
            plt.ylabel('Precision')
            plt.title(plotname)
            plt.tight_layout()

            # if save_dir is set, figure will be saved to save_dir
            if save_dir is not None:
                save_file = f"{save_dir}/{plotname.replace(' ', '_')}{save_ext}"
                print('save figure to %s' % save_file)
                fig.savefig(save_file)

            fig.canvas.draw()
            data = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8, sep='')
            data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))
            figures_np[plotname] = (data, fig)

            if show:
                plt.show()

            plt.close()

        return figures_np

    def analyze(self, save_to_dir=None, show=False, iou_threshes=None, iou_labels=None, font_size=None,
            save_ext=".pdf"):
        """Derek Hoiem et al., inspired breakdown of false positives in COCO predictions

        Usage:
        .. code-block:: python

            iou_rng = (0.5, 0.95, 0.05)  # Standard COCO evaluation metrics
            coco_eval = COCOEvalExtended(coco_true, coco_pred, iouType='bbox', iouRng=iou_rng)
            # Standard coco eval summary
            coco_eval.params.imgIds = image_ids
            coco_eval.evaluate()
            coco_eval.accumulate()
            coco_eval.summarize()
            stats = coco_eval.stats.copy()

            # Analysis plots
            coco_eval.plot_class_pr(class_names=dataset.labels, save_to_dir='{}figures'.format(name), show=True)
            iou_labels = ["C75", "C50", "LOC"]  # Standard scales for the localisation error analysis
            iou_threshes = [0.75, 0.5, 0.1]
            _ = coco_eval.analyze('{}figures'.format(name), show=True, iou_threshes=iou_threshes,
                                  iou_labels=iou_labels)


        Custom implementation for F1 support etc, originally from Matlab code for http://cocodataset.org/#detection-eval
        The categories of evaluation summarised in this plot are listed below, each step is more permissive allowing a
        clean visualisation of the errors in object detection by stacking the PR curves

        .. code-block::

            C75: PR at IoU=.75 (AP at strict IoU), area under curve corresponds to APIoU=.75 metric.
            C50: PR at IoU=.50 (AP at PASCAL IoU), area under curve corresponds to APIoU=.50 metric.
            Loc: PR at IoU=.10 (localization errors ignored, but not duplicate detections). All remaining settings use IoU=.1.
            Sim: PR after supercategory false positives (fps) are removed. Specifically, any matches to objects with a different class label but that belong to the same supercategory don't count as either a fp (or tp). Sim is computed by setting all objects in the same supercategory to have the same class label as the class in question and setting their ignore flag to 1. Note that person is a singleton supercategory so its Sim result is identical to Loc.
            Oth: PR after all class confusions are removed. Similar to Sim, except now if a detection matches any other object it is no longer a fp (or tp). Oth is computed by setting all other objects to have the same class label as the class in question and setting their ignore flag to 1.
            BG: PR after all background (and class confusion) fps are removed. For a single category, BG is a step function that is 1 until max recall is reached then drops to 0 (the curve is smoother after averaging across categories).
            FN: PR after all remaining errors are removed (trivially AP=1).


        Args:
            save_to_dir: The directory to save the plots to i.e 'figures', if None the figures are not saved
            show: Opens the plots in a matplotlib window if True
            iou_threshes: Tuple of IoU threshes to analyse (default is (0.75, 0.5 and 0.1) as per the COCO format
            iou_labels: Corresponding labels for the iou_threshes (default is ('C75', 'C50', 'LOC'))
            font_size: Display font size
            save_ext: matplotlib figure save extension (.pdf|.png|.jpg etc.)

        Returns:
            The rendered figure objects
        """
        if iou_labels is None:
            iou_labels = ['C75', 'C50', 'LOC']
        if iou_threshes is None:
            iou_threshes = [0.75, 0.5, 0.1]
        self.params.outDir = save_to_dir
        prm = copy.deepcopy(self.params)
        self.params.maxDets = [100]
        cat_ids = sorted(self.cocoGt.getCatIds())
        self.params.catIds = cat_ids
        self.params.iouThrs = np.array(iou_threshes)
        self.evaluate(cm=False)
        self.accumulate(cm=False)

        # PS storage = 0:C75, 1:C50, 2:C10/LOC, 3:SIM, 4:CLS, 5:BGR, 6:FIN
        ps = self.eval['precision']
        rc = self.eval['recall']
        ps = np.concatenate((ps, np.zeros([4] + list(ps.shape[1:]))))
        rc = np.concatenate((rc, np.zeros([4] + list(rc.shape[1:]))))

        self.params.iouThrs = np.array([0.1])
        self.params.useCats = 0

        gt = self.cocoGt
        self.analyze_figures = {}
        for k, catId in enumerate(cat_ids):
            nm = self.cocoGt.loadCats(catId)[0]
            nm = nm['name'] if 'supercategory' not in nm else nm['supercategory'] + '-' + nm['name']

            print('Analyzing %s (%d):' % (nm, k))
            start_time = time.time()

            # select detections for single category only
            self.params.keepDtCatIds = [catId]

            # compute precision but ignore superclass confusion
            cur_cat = gt.loadCats(catId)[0]
            if 'supercategory' in cur_cat:
                similar_cat_ids = gt.getCatIds(supNms=cur_cat['supercategory'])
                self.params.keepGtCatIds = similar_cat_ids
                self.params.targetCatId = catId

                # computeIoU need real catIds, we need to recover it
                self.params.catIds = cat_ids
                self.evaluate(cm=False)
                self.accumulate(cm=False)
                ps[3, :, k, :, :] = self.eval['precision'][0, :, 0, :, :]
                rc[3, k, :, :] = self.eval['recall'][0, 0, :, :]
            else:
                # skip  superclass confusion evaluation
                ps[3, :, k, :, :] = ps[2, :, k, :, :]
                rc[3, k, :, :] = rc[2, k, :, :]

            # compute precision but ignore any class confusion
            self.params.targetCatId = catId
            self.params.keepGtCatIds = cat_ids
            self.params.catIds = cat_ids
            self.evaluate(cm=False)
            self.accumulate(cm=False)
            ps[4, :, k, :, :] = self.eval['precision'][0, :, 0, :, :]
            rc[4, k, :, :] = self.eval['recall'][0, 0, :, :]

            # fill in background and false negative errors and plot
            ps[ps == -1] = 0
            ps[5, :, k, :] = (ps[4, :, k, :] > 0).astype(np.float32)
            ps[6, :, k, :] = 1
            rc[rc == -1] = 0
            rc[5, k, :] = (rc[4, k, :] > 0).astype(np.float32)
            rc[6, k, :] = 1
            end_time = time.time()

            figures = self.__makeplot(self.params.recThrs, ps[:, :, k, :, :], rc[:, k, :, :], nm,
                                      save_dir=self.params.outDir,
                                      show=show, iou_labels=iou_labels, font_size=font_size, save_ext=save_ext)
            self.analyze_figures.update(figures)
            print('Analyzing DONE (t=%0.2fs).' % (end_time - start_time))

        # reset Dt and Gt, params
        self.params = prm
        figures = self.__makeplot(self.params.recThrs, np.mean(ps, axis=2), np.mean(rc, axis=1), 'All Classes',
                                  save_dir=self.params.outDir,
                                  show=show, iou_labels=iou_labels, font_size=font_size, save_ext=save_ext)
        self.analyze_figures.update(figures)
        if 'supercategory' in list(self.cocoGt.cats.values())[0]:
            sup = [cat['supercategory'] for cat in self.cocoGt.loadCats(cat_ids)]
            print('all sup cats: %s' % (set(sup)))
            for k in set(sup):
                ps1 = np.mean(ps[:, :, np.array(sup) == k, :, :], axis=2)
                rc1 = np.mean(rc[:, np.array(sup) == k, :, :], axis=1)
                figures = self.__makeplot(self.params.recThrs, ps1, rc1, 'Overall - %s' % k,
                                          save_dir=self.params.outDir,
                                          show=show, iou_labels=iou_labels, font_size=font_size, save_ext=save_ext)
                self.analyze_figures.update(figures)
        return self.analyze_figures


class ConfusionMatrix(object):
    """Utility class for manipulating/visualising confusion matrices

        Metrics based on https://stackoverflow.com/a/48030708.
    """
    # TODO: Implement to binary confusion matrix function
    __metrics = {
        "statistics": {
            "accuracy": lambda tp, tn, fp, fn: (tp + tn) / (tp + tn + fp + fn) if (tp + tn) > 0 else 0.0,
            "f1score": lambda tp, tn, fp, fn: (2 * tp) / ((2 * tp) + (fp + fn)) if tp > 0 else 0.0,
            "sensitivity": lambda tp, tn, fp, fn: tp / (tp + fn) if tp > 0 else 0.0,
            "specificity": lambda tp, tn, fp, fn: tn / (tn + fp) if tn > 0 else 0.0,
            "precision": lambda tp, tn, fp, fn: tp / (tp + fp) if tp > 0 else 0.0,
            "recall": lambda tp, tn, fp, fn: tp / (tp + fn) if tp > 0 else 0.0,
            "tpr": lambda tp, tn, fp, fn: tp / (tp + fn) if tp > 0 else 0.0,
            "tnr": lambda tp, tn, fp, fn: tn / (tn + fp) if tn > 0 else 0.0,
            "fpr": lambda tp, tn, fp, fn: fp / (fp + tn) if fp > 0 else 0.0,
            "fnr": lambda tp, tn, fp, fn: fn / (fn + tp) if fn > 0 else 0.0,
            "ppv": lambda tp, tn, fp, fn: tp / (tp + fp) if tp > 0 else 0.0,
            "npv": lambda tp, tn, fp, fn: tn / (tn + fn) if tn > 0 else 0.0,
            "mcc": lambda tp, tn, fp, fn: ((tp * tn) - (fp * fn)) / math.sqrt(
                (tp + fp) * (tp + fn) * (tn + fp) * (tn + fn)) if (tp * tn * fn * fp) > 0 else 0.0,
            "j-statistic": lambda tp, tn, fp, fn: (tp / (tp + fn)) + (tn / (tn + fp)) - 1 if (tp > 0 or (
                    tp + fn) > 0) or (tn > 0 or (tn + fp) > 0) else -1.0
        },
        "counts": {
            "tp": lambda tp, tn, fp, fn: tp,
            "tn": lambda tp, tn, fp, fn: tn,
            "fp": lambda tp, tn, fp, fn: fp,
            "fn": lambda tp, tn, fp, fn: fn,
            "pos": lambda tp, tn, fp, fn: tp + fn,
            "neg": lambda tp, tn, fp, fn: tn + fp,
            "prop": lambda tp, tn, fp, fn: (tp + fn) / (tp + tn + fp + fn)
        }
    }

    def __init__(self, confusion_matrix, classes=None, class_names=None):
        self.matrix = confusion_matrix
        if classes is None:
            classes = [i for i in range(confusion_matrix.shape[0])]
        if class_names is None:
            class_names = [str(i) for i in classes]
        self.classes = classes
        self.class_names = class_names
        self.class2index = {key: i for i, key in enumerate(self.classes)}
        self.index2class = {i: key for i, key in enumerate(self.classes)}
        self.metrics = self.__metrics

        self.matrix_sum = sum([sum(row) for row in self.matrix])
        self.normalised_matrix = [row for row in
                                  map(lambda i: list(map(lambda j: j / self.matrix_sum, i)), self.matrix)]
        self.results = None
        self.__compute_stats()

    @staticmethod
    def get_matrix(actual, predicted, labels=None):
        if labels is None:
            labels = actual
        classes = sorted(set(labels))
        matrix = np.zeros((len(classes), len(classes)))
        class2index = {key: i for i, key in enumerate(classes)}
        index2class = {i: key for i, key in enumerate(classes)}
        for pv, av in zip(predicted, actual):
            matrix[class2index[pv]][class2index[av]] += 1
        return matrix

    def __get_class_balance(self):
        return {
            self.index2class[i]: self.results[self.index2class[i]]["counts"]["pos"] for i, _ in enumerate(self.classes)
        }

    def __get_cohen_kappa(self, weight=None):
        """Based on scikit-learn metrics.cohen_kappa"""
        n_classes = self.matrix.shape[0]
        sum0 = np.sum(self.matrix, axis=0)
        sum1 = np.sum(self.matrix, axis=1)
        expected = np.outer(sum0, sum1) / np.sum(sum0)

        weights = None
        if weights is None:
            w_mat = np.ones([n_classes, n_classes], dtype=np.int)
            w_mat.flat[:: n_classes + 1] = 0
        elif weights == "linear" or weights == "quadratic":
            w_mat = np.zeros([n_classes, n_classes], dtype=np.int)
            w_mat += np.arange(n_classes)
            if weights == "linear":
                w_mat = np.abs(w_mat - w_mat.T)
            else:
                w_mat = (w_mat - w_mat.T) ** 2
        else:
            raise ValueError("Unknown kappa weighting type.")

        k = np.sum(w_mat * self.matrix) / np.sum(w_mat * expected)
        return 1 - k

    def __compute_stats(self):
        self.results = OrderedDict(((c, {
            "counts": OrderedDict(),
            "stats": OrderedDict()
        }) for c in self.classes))
        for i in range(len(self.classes)):
            row = sum(self.matrix[i][:])
            col = sum([row[i] for row in self.matrix])
            tp = self.matrix[i][i]
            fp = row - tp
            fn = col - tp
            tn = self.matrix_sum - row - col + tp
            for count, func in self.metrics["counts"].items():
                self.results[self.index2class[i]]["counts"][count] = self.metrics["counts"][count](tp, tn, fp, fn)
            for stat, func in self.metrics["statistics"].items():
                self.results[self.index2class[i]]["stats"][stat] = self.metrics["statistics"][stat](tp, tn, fp, fn)

    def summarise(self):
        first_column_width = max(7, max([len(s) for s in self.class_names]))
        count_metrics_len = len(self.metrics["counts"].keys())
        count_type_max_len = max([len(s) + 1 for s in self.metrics["counts"].keys()])
        counts_fmt = f" {{:<{first_column_width}}}  " + ' '.join([f"{{:<{count_type_max_len}}}"] * count_metrics_len)
        counts_title = counts_fmt.format(*(["Counts"] + list(self.metrics["counts"].keys())))
        stat_metrics_len = len(self.metrics["statistics"].keys())
        stat_type_max_len = max([len(s) + 1 for s in self.metrics["statistics"].keys()])
        stats_fmt = f" {{:<{first_column_width}}}  " + ' '.join([f"{{:<{stat_type_max_len}}}"] * stat_metrics_len)
        stats_title = stats_fmt.format(*(["Stats"] + list(self.metrics["statistics"].keys())))

        print(counts_title)
        for class_idx, class_name in enumerate(self.class_names):
            row_fmt = f" {{:<{first_column_width}}}  " + ' '.join([f"{{:<{count_type_max_len}}}"] * count_metrics_len)
            count_values = list(self.results[self.index2class[class_idx]]["counts"].values())
            row_str = row_fmt.format(*([class_name] + [
                ("{:d}" if r.is_integer() else "{:0.2f}%").format(
                    int(r) if r.is_integer() else r * 100) for r in count_values]))
            print(row_str)

        print()
        print(stats_title)
        for class_idx, class_name in enumerate(self.class_names):
            row_fmt = f" {{:<{first_column_width}}}  " + ' '.join([f"{{:<{stat_type_max_len}}}"] * stat_metrics_len)
            stat_values = list(self.results[self.index2class[class_idx]]["stats"].values())
            row_str = row_fmt.format(*([class_name] + [
                ("{:d}" if r.is_integer() else "{:0.2f}%").format(
                    int(r) if r.is_integer() else r * 100) for r in stat_values]))
            print(row_str)

        print((f"\n {{:<{first_column_width}}}  ").format("Cohen Kappa") + (f"{{:<{stat_type_max_len}}}\n").format(
            "{:0.2f}".format(self.__get_cohen_kappa())
        ))

    def tabulate(self, normalised=False):
        first_column_width = max(9, max([len(s) for s in self.class_names]) + 1)
        count_width = max([len(str(c)) for c, n in self.__get_class_balance().items()])
        column_widths = max(first_column_width, count_width)
        title_fmt = f" {{:<{first_column_width}}} " + ' '.join([f"{{:<{column_widths}}}"] * len(self.class_names))
        title = title_fmt.format(*(["Actual"] + [c + "ᴬ" for c in self.class_names]))
        print(title)
        print((f" {{:<{first_column_width}}} ").format("Predicted"))
        matrix = self.normalised_matrix if normalised else self.matrix

        for class_name, row in zip(self.class_names, matrix):
            row_fmt = f" {{:<{first_column_width}}} " + ' '.join([f"{{:<{column_widths}}}"] * len(self.class_names))
            row_str = row_fmt.format(*([class_name + "ᴾ"] + [
                ("{:d}" if r.is_integer() else "{:0.2f}%").format(
                    int(r) if r.is_integer() else r * 100) for r in row
            ]))
            print(row_str)

        print()



from typing import List, Tuple

import numpy as np


def get_heatmap_from_points(image_shape: Tuple[int, int], points: List[Tuple[int, int]]):
    points = np.array(points)
    yy, xx = np.meshgrid(np.arange(image_shape[0]), np.arange(image_shape[1]))
    xd = np.array([(xx - points[i, 1]) ** 2 for i in range(len(points))])
    yd = np.array([(yy - points[i, 0]) ** 2 for i in range(len(points))])
    d = np.min(xd + yd, axis=0) ** 0.5
    return d / d.ptp()


def __main():
    pass


if __name__ == '__main__':
    __main()
